#
# Copyright (C) 2020 Sylvain Marsat.
#
#

import json
import numpy as np
import pymultinest

import lisabeta
import lisabeta.tools.pytools as pytools
import lisabeta.tools.pyspline as pyspline
import lisabeta.lvk.pyLVKresponse as pyLVKresponse
import lisabeta.lvk.lvk as lvk

import argparse

parser = argparse.ArgumentParser(description='Run LVK parameter inference \
                                              using lisabeta and pymultinest.')

parser.add_argument('input_file', type=str, help='Input json file \
                                                  with all parameters.')
args = parser.parse_args()

# Full list of physical params - infer_params are a subset
# TODO: generalize
listparams = [
    "Mchirp",
    "eta",
    "chi1",
    "chi2",
    "Deltat",
    "dist",
    "inc",
    "phi",
    "ra",
    "dec",
    "psi"]

# Default waveform params
waveform_params_default = {
    "minf": 10.,
    "maxf": 2048.,
    "fstart": None,
    "fend": None,
    "tstart": None,
    "tend": None,
    "phiref": 0.0,
    "fref_for_phiref": 0.0,
    "tref": 0.0,
    "fref_for_tref": 0.0,
    "force_phiref_fref": True,
    "toffset": 0.0,
    "acc": 1e-4,
    "order_fresnel_stencil": 0,
    "approximant": "IMRPhenomD",
    "detectors": ["LHO", "LLO", "VIRGO"]
}

# Default pymultinest params
run_params_musthave = ['outputfiles_basename']
run_params_default = {
    "lisabeta_data_dir": lvk.default_lisabeta_data_dir,
    "zerolike": False,
    "n_live_points": 1000,
    "evidence_tolerance": 0.1,
    "sampling_efficiency": 0.5,
    "seed": -1,
    "resume": False,
    "max_iter": 0,
    "verbose": False,
    "write_output": True,
    "n_iter_before_update": 100,
    "importance_nested_sampling": False,
    "multimodal": False,
    "const_efficiency_mode": False
}

# Default parameter range for angular parameters
params_range_default = {
    "inc": [0., np.pi],
    "phi": [-np.pi, np.pi],
    "ra": [-np.pi, np.pi],
    "dec": [-np.pi/2, np.pi/2],
    "psi": [0., np.pi]
}

# Utility to go from Mchirp-eta variables to m1-m2
def convert_params_m1m2(params_Mchirpeta):
    params_m1m2 = params_Mchirpeta.copy()
    Mchirp = params_m1m2.pop('Mchirp')
    eta = params_m1m2.pop('eta')
    m1 = pytools.m1ofMchirpeta(Mchirp, eta)
    m2 = pytools.m2ofMchirpeta(Mchirp, eta)
    params_m1m2['m1'] = m1
    params_m1m2['m2'] = m2
    return params_m1m2
# Utility to go from m1-m2 variables to Mchirp-eta
def convert_params_Mchirpeta(params_m1m2):
    params_Mchirpeta = params_m1m2.copy()
    m1 = params_Mchirpeta.pop('m1')
    m2 = params_Mchirpeta.pop('m2')
    Mchirp = pytools.Mchirpofm1m2(m1, m2)
    eta = pytools.etaofm1m2(m1, m2)
    params_Mchirpeta['Mchirp'] = Mchirp
    params_Mchirpeta['eta'] = eta
    return params_Mchirpeta

# Utilities for cube-to-param mapping
def map_cube_to_param(priortype, paramrange, x):
    if priortype=='uniform':
        return paramrange[0] + x * (paramrange[1] - paramrange[0])
    elif priortype=='sin': # NOTE: cos decreasing on [0,pi]
        cosrange = [np.cos(paramrange[1]), np.cos(paramrange[0])]
        return np.arccos(cosrange[0] + x * (cosrange[1] - cosrange[0]))
    elif priortype=='cos': # NOTE: sin increasing on [-pi/2,pi/2]
        sinrange = [np.sin(paramrange[0]), np.sin(paramrange[1])]
        return np.arcsin(sinrange[0] + x * (sinrange[1] - sinrange[0]))
    else:
        raise ValueError('priortype not recognized.')

if __name__ == '__main__':

    # Load json file with all parameters
    with open(args.input_file, 'r') as input_file:
        input_params = json.load(input_file)
    # Source params and prior params must be given
    source_params = input_params['source_params']
    prior_params = input_params['prior_params']
    # Waveform params and run params have predefined default values
    waveform_params = waveform_params_default.copy()
    waveform_params.update(input_params['waveform_params'])
    run_params = run_params_default.copy()
    run_params.update(input_params['run_params'])
    zerolike = run_params.pop('zerolike', False)
    lisabeta_data_dir = run_params.pop('lisabeta_data_dir', lvk.default_lisabeta_data_dir)
    # run params must contain output prefix for files -- no default here
    for musthave in run_params_musthave:
        if not (musthave in run_params):
            raise ValueError('run_params must contain %s.' % (musthave))

    # Use default param ranges to complete input ranges -- check completeness
    prior_type = prior_params['prior_type']
    params_range = prior_params['params_range']
    infer_params = prior_params['infer_params']
    for i, param in enumerate(infer_params):
        if params_range[i] == []:
            params_range[i] = params_range_default[param]
    if [] in params_range:
        raise ValueError('Some parameter ranges are neither given \
                      nor covered by the defaults.')

    # Load and interpolate noise data
    noises = {}
    for det in waveform_params['detectors']:
        noisedata = np.loadtxt(lisabeta_data_dir + lvk.noise_data_files[det])
        noise_f = np.copy(noisedata[:,0])
        noise_Sn = np.copy(noisedata[:,1]**2) # sqrt(Sn) in files
        noisesplineClass = pyspline.CubicSpline(noise_f, noise_Sn)
        noises[det] = noisesplineClass.get_spline()

    # Generate injection signal
    signalground_inj = lvk.GenerateLVKSignal(source_params, noises,
                                             **waveform_params)

    # Parameters inferred and parameters fixed
    infer_params = prior_params['infer_params']
    fixed_params = [p for p in listparams if p not in infer_params]
    n_dim = len(infer_params)
    n_fixed = len(fixed_params)
    # NOTE: should be able to set n_params and carry along extra fixed params
    # but did not succeed: was getting undefined behaviour in outputs
    n_params = n_dim

    # Wrapped parameters
    wrapped_params = [0] * n_dim
    for i in range(n_dim):
        wrapped_params[i] = int(prior_params['wrap_params'][i])
    run_params['wrapped_params'] = wrapped_params

    # Define function mapping unit cube to physical params according to priors
    # NOTE: this function is in-place for cube and does not return
    def cubeparams_to_physparams(cube, ndim, nparams):
        assert nparams==n_params and ndim==n_dim
        for i in range(n_dim):
            cube[i] = map_cube_to_param(prior_type[i], params_range[i], cube[i])

    # Define loglikelihood function
    # NOTE: here cube contains already physical parameters, not unit cube values
    def loglike(cube, ndim, nparams):
        assert nparams==n_params and ndim==n_dim
        if zerolike: return 0.
        template_params = {}
        for i in range(n_dim):
            template_params[infer_params[i]] = cube[i]
        for i in range(n_fixed):
            template_params[fixed_params[i]] = \
                                    source_params_Mchirpeta[fixed_params[i]]
        #print(template_params)
        template_params_m1m2 = convert_params_m1m2(template_params)
        lnL = lvk.LogLikelihoodLVK(template_params_m1m2, signalground_inj,
                                   noises, **waveform_params)
        #print(lnL)
        return lnL

    # A la grace de dieu
    pymultinest.run(loglike, cubeparams_to_physparams, n_dim, n_params=n_params,
                    **run_params)



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython wrapping of C spline interpolation tools.
"""

import numpy as np
cimport numpy as np

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *


cdef class CubicSpline:
    """ Cubic spline with the not-a-knot boundary condition, as a matrix.
    """

    cdef real_matrix* Cspline   # pointer to C spline matrix
    cdef real_matrix* Cspline_d # pointer to C spline derivative matrix
    cdef real_matrix* Cspline_dd # pointer to C spline double derivative matrix
    cdef real_matrix* Cspline_int # pointer to C spline integral matrix
    cdef real_vector* Cx        # pointer to C x vector
    cdef real_vector* Cy        # pointer to C y vector
    cdef x, y                   # input numpy arrays
    cdef spline                 # numpy array for spline matrix
    cdef spline_d               # numpy array for spline derivative matrix
    cdef spline_dd              # numpy array for spline double derivative matrix
    cdef spline_int             # numpy array for spline integral matrix (4th order)


    def __init__(self,
                 np.ndarray[np.float_t, ndim=1] x,
                 np.ndarray[np.float_t, ndim=1] y):
        """Constructor
        Args:
          x                     # Input 1D numpy array with x values
          y                     # Input 1D numpy array with y values
        Keyword args:
          None
        """
        self.x = x
        self.y = y

        self.Cspline = NULL
        self.Cspline_d = NULL
        self.Cspline_dd = NULL
        self.Cspline_int = NULL
        cdef double val

        # If the input is an empty array, init to empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(x)==0 or len(y)==0:
            self.spline = np.empty((0,0), dtype=float)
            return

        # Build a real_vector representation of the input numpy arrays
        if not x.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array x is not C_CONTIGUOUS')
        if not y.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array y is not C_CONTIGUOUS')
        cdef double* x_data = <double *> &x[0]
        cdef double* y_data = <double *> &y[0]
        self.Cx = real_vector_view(x_data, x.shape[0])
        self.Cy = real_vector_view(y_data, y.shape[0])

        if len(x)>3:
            ret = BuildNotAKnotSpline(
                &self.Cspline,
                self.Cx,
                self.Cy,
            );
            if ret == _FAILURE:
                raise ValueError("Call to BuildNotAKnotSpline() failed.")
        else:
            #Cannot BuildNotAKnotSpline for n<4
            #Instead we build a QuadSpline and use its coefficients
            quadspline=QuadSpline(x,y).get_spline()

            self.Cspline=real_matrix_alloc(len(x),5)
            for i in range(len(x)): #Not very optimized, but <16 array elems
                for j in range(5):
                    if j<4:
                        val=quadspline[i,j]
                    else:val=0
                    self.Cspline.data[i * 5 + j] = val

        # Cast C double array to numpy via a MemoryView
        cdef int m = self.Cspline.size1
        cdef int n = self.Cspline.size2
        cdef double[::1] view_Cspline = \
          <(double)[:(m * n)]> self.Cspline.data
        self.spline = np.reshape(np.asarray(view_Cspline), (m, n))
        #self.spline = real_matrix_to_np_array(self.Cspline)

    def __dealloc__(self):
        """Destructor
        """
        if self.Cspline != NULL:
            real_matrix_free(self.Cspline)
        if self.Cspline_d != NULL:
            real_matrix_free(self.Cspline_d)
        if self.Cspline_dd != NULL:
            real_matrix_free(self.Cspline_dd)
        if self.Cspline_int != NULL:
            real_matrix_free(self.Cspline_int)
        if self.Cx != NULL:
            real_vector_view_free(self.Cx)
        if self.Cy != NULL:
            real_vector_view_free(self.Cy)

    def get_spline(self):
        return np.copy(self.spline)

    def get_spline_d(self):

        # If the spline is an empty array (or len=1 where deriv is undefined), return empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(self.spline)==0 or len(self.x)<2:
            return np.empty((0,0), dtype=float)

        if self.Cspline_d == NULL:
            ret = CubicSplineDerivative(
                &self.Cspline_d,
                self.Cspline
            );
            if ret == _FAILURE:
                raise ValueError("Call to CubicSplineDerivative() failed.")

        # Cast C double array to numpy via a MemoryView
        #self.spline_d = real_matrix_to_np_array(self.Cspline_d)
        cdef int m = self.Cspline_d.size1
        cdef int n = self.Cspline_d.size2
        cdef double[::1] view_Cspline_d = \
          <(double)[:(m * n)]> self.Cspline_d.data
        self.spline_d = np.reshape(np.asarray(view_Cspline_d), (m, n))

        return np.copy(self.spline_d)

    def get_spline_dd(self):

        # If the spline is an empty array, return empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(self.spline)==0:
            return np.empty((0,0), dtype=float)

        if self.Cspline_dd == NULL:
            ret = CubicSplineDoubleDerivative(
                &self.Cspline_dd,
                self.Cspline
            );
            if ret == _FAILURE:
                raise ValueError("Call to CubicSplineDerivative() failed.")

        # Cast C double array to numpy via a MemoryView
        #self.spline_d = real_matrix_to_np_array(self.Cspline_d)
        cdef int m = self.Cspline_dd.size1
        cdef int n = self.Cspline_dd.size2
        cdef double[::1] view_Cspline_dd = \
          <(double)[:(m * n)]> self.Cspline_dd.data
        self.spline_dd = np.reshape(np.asarray(view_Cspline_dd), (m, n))

        return np.copy(self.spline_dd)

    def get_spline_int(self):

        # If the spline is an empty array, return empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(self.spline)==0:
            return np.empty((0,0), dtype=float)

        if self.Cspline_int == NULL:
            ret = CubicSplineIntegral(
                &self.Cspline_int,
                self.Cspline
            );
            if ret == _FAILURE:
                raise ValueError("Call to CubicSplineDerivative() failed.")

        # Cast C double array to numpy via a MemoryView
        #self.spline_int = real_matrix_to_np_array(self.Cspline_int)
        cdef int m = self.Cspline_int.size1
        cdef int n = self.Cspline_int.size2
        cdef double[::1] view_Cspline_int = \
          <(double)[:(m * n)]> self.Cspline_int.data
        self.spline_int = np.reshape(np.asarray(view_Cspline_int), (m, n))

        return np.copy(self.spline_int)

cdef class QuadSpline:
    """ Quadratic spline, as a matrix.
    """

    cdef real_matrix* Cspline   # pointer to C spline matrix
    cdef real_matrix* Cspline_d # pointer to C spline derivative matrix
    # cdef real_matrix* Cspline_int # pointer to C spline integral matrix
    cdef real_vector* Cx        # pointer to C x vector
    cdef real_vector* Cy        # pointer to C y vector
    cdef x, y                   # input numpy arrays
    cdef spline                 # numpy array for spline matrix
    cdef spline_d               # numpy array for spline derivative matrix
    cdef spline_int             # numpy array for spline integral matrix (4th order)


    def __init__(self,
                 np.ndarray[np.float_t, ndim=1] x,
                 np.ndarray[np.float_t, ndim=1] y):
        """Constructor
        Args:
          x                     # Input 1D numpy array with x values
          y                     # Input 1D numpy array with y values
        Keyword args:
          None
        """
        self.x = x
        self.y = y

        self.Cspline = NULL
        self.Cspline_d = NULL
        cdef double slope

        # self.Cspline_int = NULL

        # If the input is an empty array, init to empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(x)==0 or len(y)==0:
            self.spline = np.empty((0,0), dtype=float)
            return

        # Build a real_vector representation of the input numpy arrays
        if not x.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array x is not C_CONTIGUOUS')
        if not y.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array y is not C_CONTIGUOUS')
        cdef double* x_data = <double *> &x[0]
        cdef double* y_data = <double *> &y[0]
        self.Cx = real_vector_view(x_data, x.shape[0])
        self.Cy = real_vector_view(y_data, y.shape[0])

        if len(x)>2:
            ret = BuildQuadSpline(
                &self.Cspline,
                self.Cx,
                self.Cy,
            );
            if ret == _FAILURE:
                raise ValueError("Call to BuildNotAKnotSpline() failed.")
        else:
            #Cannot QuadSpline for n<3
            #Instead we set coefficients for linear interpolation, or constant
            slope=0;
            if len(x)>1: slope=(y_data[1]-y_data[0])/(x_data[1]-x_data[0])
            self.Cspline=real_matrix_alloc(len(x),4)
            for i in range(len(x)): #Not very optimized, but <16 array elems
                self.Cspline.data[i * 4 + 0] = x_data[i]
                self.Cspline.data[i * 4 + 1] = y_data[i]
                self.Cspline.data[i * 4 + 2] = slope
                self.Cspline.data[i * 4 + 3] = 0

        # Cast C double array to numpy via a MemoryView
        cdef int m = self.Cspline.size1
        cdef int n = self.Cspline.size2
        cdef double[::1] view_Cspline = \
          <(double)[:(m * n)]> self.Cspline.data
        self.spline = np.reshape(np.asarray(view_Cspline), (m, n))
        #self.spline = real_matrix_to_np_array(self.Cspline)

    def __dealloc__(self):
        """Destructor
        """
        if self.Cspline != NULL:
            real_matrix_free(self.Cspline)
        if self.Cspline_d != NULL:
            real_matrix_free(self.Cspline_d)
        # if self.Cspline_int != NULL:
        #     real_matrix_free(self.Cspline_int)
        if self.Cx != NULL:
            real_vector_view_free(self.Cx)
        if self.Cy != NULL:
            real_vector_view_free(self.Cy)

    def get_spline(self):
        return np.copy(self.spline)

    def get_spline_d(self):

        # If the spline is an empty array (or len=1 where deriv is undefined), return empty array and stop there
        # NOTE: __init__ is required to return nothing
        if len(self.spline)==0 or len(self.x)<2:
            return np.empty((0,0), dtype=float)

        if self.Cspline_d == NULL:
            ret = QuadSplineDerivative(
                &self.Cspline_d,
                self.Cspline
            );
            if ret == _FAILURE:
                raise ValueError("Call to CubicSplineDerivative() failed.")

        # Cast C double array to numpy via a MemoryView
        #self.spline_d = real_matrix_to_np_array(self.Cspline_d)
        cdef int m = self.Cspline_d.size1
        cdef int n = self.Cspline_d.size2
        cdef double[::1] view_Cspline_d = \
          <(double)[:(m * n)]> self.Cspline_d.data
        self.spline_d = np.reshape(np.asarray(view_Cspline_d), (m, n))

        return np.copy(self.spline_d)

    # def get_spline_int(self):
    #
    #     if self.Cspline_int == NULL:
    #         ret = CubicSplineIntegral(
    #             &self.Cspline_int,
    #             self.Cspline
    #         );
    #         if ret == _FAILURE:
    #             raise ValueError("Call to CubicSplineDerivative() failed.")
    #
    #     # Cast C double array to numpy via a MemoryView
    #     #self.spline_int = real_matrix_to_np_array(self.Cspline_int)
    #     cdef int m = self.Cspline_int.size1
    #     cdef int n = self.Cspline_int.size2
    #     cdef double[::1] view_Cspline_int = \
    #       <(double)[:(m * n)]> self.Cspline_int.data
    #     self.spline_int = np.reshape(np.asarray(view_Cspline_int), (m, n))
    #
    #     return np.copy(self.spline_int)

def spline_eval_vector(np.ndarray[np.float_t, ndim=2] spline,
                       np.ndarray[np.float_t, ndim=1] xvec,
                       extrapol_zero=False):

    # If the input is an empty array, return empty array and stop there
    # NOTE: __init__ is required to return nothing
    if len(spline)==0 or len(xvec)==0:
        return np.empty(0, dtype=float)

    # Build real_vector representation of input numpy array for x
    cdef double* xvec_data = <double *> &xvec[0]
    Cxvec = real_vector_view(xvec_data, xvec.shape[0])

    # Build real_matrix representation of input numpy array for spline
    if not spline.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array spline is not C_CONTIGUOUS.')
    if not spline.shape[1] == 5:
        raise ValueError(
        'Input array for spline has dimensions (%d, %d) instead of (n,5)',
        spline.shape[0], spline.shape[1])
    if not xvec.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array xvec is not C_CONTIGUOUS')
    cdef real_matrix* Cspline = NULL
    cdef double* spline_data = <double *> &spline[0,0]
    Cspline = real_matrix_view(spline_data,
                               spline.shape[0], spline.shape[1])

    cdef real_vector* Cyvec = NULL
    ret = EvalCubicSplineOnVector(
        &Cyvec,
        Cxvec,
        Cspline,
        extrapol_zero
    );
    if ret == _FAILURE:
        raise ValueError("Call to EvalCubicSplineOnVector() failed.")

    # Cast C double array to numpy via a MemoryView
    # NOTE: np.copy to return an array independent of C allocated memory
    cdef int n = Cyvec.size
    cdef double[::1] view_Cyvec = <(double)[:n]> Cyvec.data
    yvec = np.copy(np.asarray(view_Cyvec))

    # Cleanup
    if Cyvec != NULL:
        real_vector_free(Cyvec)
    if Cxvec != NULL:
        real_vector_view_free(Cxvec)
    if Cspline != NULL:
        real_matrix_view_free(Cspline)

    return yvec

def quadspline_eval_vector(np.ndarray[np.float_t, ndim=2] spline,
                           np.ndarray[np.float_t, ndim=1] xvec,
                           extrapol_zero=False):

    # If the input is an empty array, return empty array and stop there
    # NOTE: __init__ is required to return nothing
    if len(spline)==0 or len(xvec)==0:
        return np.empty(0, dtype=float)

    # Build real_vector representation of input numpy array for x
    cdef double* xvec_data = <double *> &xvec[0]
    Cxvec = real_vector_view(xvec_data, xvec.shape[0])

    # Build real_matrix representation of input numpy array for spline
    if not spline.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array quad spline is not C_CONTIGUOUS.')
    if not spline.shape[1] == 4:
        raise ValueError(
        'Input array for  quad spline has dimensions (%d, %d) instead of (n,4)',
        spline.shape[0], spline.shape[1])
    if not xvec.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array xvec is not C_CONTIGUOUS')
    cdef real_matrix* Cspline = NULL
    cdef double* spline_data = <double *> &spline[0,0]
    Cspline = real_matrix_view(spline_data,
                               spline.shape[0], spline.shape[1])

    cdef real_vector* Cyvec = NULL
    ret = EvalQuadSplineOnVector(
        &Cyvec,
        Cxvec,
        Cspline,
        extrapol_zero
    );
    if ret == _FAILURE:
        raise ValueError("Call to EvalQuadSplineOnVector() failed.")

    # Cast C double array to numpy via a MemoryView
    # NOTE: np.copy to return an array independent of C allocated memory
    cdef int n = Cyvec.size
    cdef double[::1] view_Cyvec = <(double)[:n]> Cyvec.data
    yvec = np.copy(np.asarray(view_Cyvec))

    # Cleanup
    if Cyvec != NULL:
        real_vector_free(Cyvec)
    if Cxvec != NULL:
        real_vector_view_free(Cxvec)
    if Cspline != NULL:
        real_matrix_view_free(Cspline)

    return yvec

cpdef np.ndarray[np.float_t, ndim=1] resample(np.ndarray[np.float_t, ndim=1] xnew, np.ndarray[np.float_t, ndim=1] xold,y):
    """
    Resample the spline function on the xnew grid.  For values outside the range of the spline data fill with zeros.
    """

    #cdef np.ndarray[np.float_t, ndim=1] ynew
    cdef np.ndarray[np.float_t, ndim=2] yspline
    ynew=np.zeros(len(xnew),dtype=y.dtype)
    #Figure out the regions not to pad with zero
    cdef double xmin,xmax,newxmin,newxmax
    xmin=xold[0]
    xmax=xold[-1]
    newxmin=xnew[0]
    newxmax=xnew[-1]
    if newxmin>xmax or newxmax<xmin: return ynew
    if newxmin<xmin:imin=np.argmax(xnew>xmin)
    else: imin=0
    if newxmax>xmax:iend=imin+np.argmax(xnew[imin:]>xmax)
    else: iend=len(xnew)
    #Construct splines and fill
    if y.dtype==np.cdouble:
        yspline=CubicSpline(xold,y.real.copy()).get_spline()
        ynew[imin:iend]=spline_eval_vector(yspline,xnew[imin:iend])
        yspline=CubicSpline(xold,y.imag.copy()).get_spline()
        ynew[imin:iend]=ynew[imin:iend]+1j*spline_eval_vector(yspline,xnew[imin:iend])
    else:
        yspline=CubicSpline(xold,y).get_spline()
        ynew[imin:iend]=spline_eval_vector(yspline,xnew[imin:iend])
    return ynew

# UNFINISHED
# REQUIRES DEBUGGING
# cpdef np.ndarray[np.complex_t, ndim=1] eval_cexp_cubic_spline_constdeltax(
#                                     np.ndarray[np.float_t, ndim=2] phase_spline,
#                                     np.ndarray[np.float_t, ndim=1] xvec):
#     """
#     Evaluate exp[i*phi] on an input array of x's
#     with phi given by a cubic spline.
#     Args:
#       phi             # Input 2D (5 cols) numpy array for the cubic phase spline
#       x               # Input 1D numpy array of x values, evenly spaced
#     Keyword args:
#       None
#     """
#
#     # Build real_vector representation of input numpy array for x
#     cdef double* xvec_data = <double *> &xvec[0]
#     Cxvec = real_vector_view(xvec_data, xvec.shape[0])
#
#     # Build real_matrix representation of input numpy array for spline
#     if not phase_spline.flags['C_CONTIGUOUS']:
#         raise ValueError('Input numpy array spline is not C_CONTIGUOUS.')
#     if not phase_spline.shape[1] == 5:
#         raise ValueError(
#         'Input array for spline has dimensions (%d, %d) instead of (n,5)',
#         phase_spline.shape[0], phase_spline.shape[1])
#     if not xvec.flags['C_CONTIGUOUS']:
#         raise ValueError('Input numpy array xvec is not C_CONTIGUOUS')
#     cdef real_matrix* Cspline = NULL
#     cdef double* phase_spline_data = <double *> &phase_spline[0,0]
#     Cspline = real_matrix_view(phase_spline_data,
#                                phase_spline.shape[0], phase_spline.shape[1])
#
#     cdef complex_vector* Ccexpvec = NULL
#     ret = EvalCubicSplineCExpOnConstDeltaxVector(
#         &Ccexpvec,
#         Cxvec,
#         Cspline
#     );
#     if ret == _FAILURE:
#         raise ValueError("Call to EvalCubicSplineCExpOnVector() failed.")
#
#     # Cast C double array to numpy via a MemoryView
#     # NOTE: np.copy to return an array independent of C allocated memory
#     cdef int n = Ccexpvec.size
#     cdef double complex[::1] view_Ccexpvec = <(double complex)[:n]> Ccexpvec.data
#     cexpvec = np.copy(np.asarray(view_Ccexpvec))
#
#     # Cleanup
#     if Ccexpvec != NULL:
#         complex_vector_free(Ccexpvec)
#     if Cxvec != NULL:
#         real_vector_view_free(Cxvec)
#     if Cspline != NULL:
#         real_matrix_view_free(Cspline)
#
#     return cexpvec



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython wrapping of C generic waveforms tools, python tools.
"""


import numpy as np
cimport numpy as np

import h5py

import lisabeta.pyconstants as pyconstants
import lisabeta.tools.pyspline as pyspline
from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *

from scipy.interpolate import InterpolatedUnivariateSpline as spline

################################################################################
# Functions to manipulate physical parameters
################################################################################

# List of mass and spin parameters
list_mass_params = ['m1', 'm2', 'M', 'q', 'Mchirp', 'eta']
list_spin_params = ['chi1', 'chi2', 'chip', 'chim', 'chis', 'chia', 'chiPN']

# Physical check on inntrinsic params
def check_physical_params(params):
    return np.all(np.array([
        params['m1'] > 0.,
        params['m2'] > 0.,
        params['M'] > 0.,
        params['q'] >= 1.,
        params['Mchirp'] > 0.,
        params['eta'] > 0.,
        params['eta'] <= 0.25,
        params['m1'] >= params['m2'],
        params['chi1'] >= -1.,
        params['chi1'] <= 1.,
        params['chi2'] >= -1.,
        params['chi2'] <= 1.,
        params['chip'] >= -1.,
        params['chip'] <= 1.,
        params['chim'] >= -1.,
        params['chim'] <= 1.,
        params['chis'] >= -1.,
        params['chis'] <= 1.,
        params['chia'] >= -1.,
        params['chia'] <= 1.,
        params['chiPN'] >= -1.,
        params['chiPN'] <= 1.,
    ]))

# Conversion between  mass ratio and symmetric mass ratio
def qofeta(eta):
    return (1.0 + np.sqrt(1.0 - 4.0*eta) - 2.0*eta) / (2.0*eta)
def etaofq(q):
    return q/(1.0 + q)**2
# Mass difference delta = (m1-m2) / (m1+m2)
def deltaofq(q):
    return (q-1.) / (q+1.)

# Conversion between mass parameters
def Mchirpofm1m2(m1, m2):
    return (m1*m2)**(3./5) / (m1+m2)**(1./5)
def etaofm1m2(m1, m2):
    return (m1*m2)/(m1+m2)**2.
def m1ofMchirpeta(Mchirp, eta):
    M = Mchirp * np.power(eta, -3./5)
    delta = np.sqrt(1 - 4*eta)
    return M * (1+delta)/2.
def m2ofMchirpeta(Mchirp, eta):
    M = Mchirp * np.power(eta, -3./5)
    delta = np.sqrt(1 - 4*eta)
    return M * (1-delta)/2.
def Mofm1m2(m1, m2):
    return m1 + m2
def qofm1m2(m1, m2):
    return m1/m2
def m1ofMq(M, q):
    return M * q/(1 + q)
def m2ofMq(M, q):
    return M * 1/(1 + q)
def MofMchirpeta(Mchirp, eta):
    return Mchirp * np.power(eta, -3./5)
def qofMchirpeta(Mchirp, eta):
    return qofeta(eta)
def MchirpofMq(M, q):
    eta = etaofq(q)
    return M * np.power(eta, 3./5)
def etaofMq(M, q):
    return etaofq(q)

# Conversion between spin parameters
def chisofchi1chi2(chi1, chi2):
    return (chi1 + chi2) / 2.
def chiaofchi1chi2(chi1, chi2):
    return (chi1 - chi2) / 2.
def chi1ofchischia(chis, chia):
    return chis + chia
def chi2ofchischia(chis, chia):
    return chis - chia
def chipofchi1chi2(chi1, chi2, q):
    return q / (1.+q) * chi1 + 1 / (1.+q) * chi2
def chimofchi1chi2(chi1, chi2, q):
    return q / (1.+q) * chi1 - 1 / (1.+q) * chi2
def chi1ofchipchim(chip, chim, q):
    return (1.+q) / (2*q) * (chip + chim)
def chi2ofchipchim(chip, chim, q):
    return (1.+q) / (2) * (chip - chim)
def chipofchischia(chis, chia, q):
    return chis + (q-1) / (1.+q) * chia
def chimofchischia(chis, chia, q):
    return chia + (q-1) / (1.+q) * chis
def chisofchipchim(chip, chim, q):
    return (1+q)**2/(4.*q) * (chip - (q-1) / (1.+q) * chim)
def chiaofchipchim(chip, chim, q):
    return (1+q)**2/(4.*q) * (chim - (q-1) / (1.+q) * chip)
def chiPNofchi1chi2(chi1, chi2, q):
    return 1./113 * etaofq(q) * ((113*q + 75.)*chi1 + (113/q + 75.)*chi2)
def chiPNofchipchim(chip, chim, q):
    return 1./113 * (94*chip + (q-1.)/(q+1.)*19*chim)
def chiPNofchischia(chis, chia, q):
    return 1./113 * etaofq(q) * ((113*q + 113/q + 150.)*chis + (113*q - 113/q)*chia)
def chipofchiPNchim(chiPN, chim, q):
    return 1./94 * (113*chiPN - (q-1.)/(q+1.)*19*chim)

# Complete mass parameters: from any set (m1,m2), (M,q), (Mchirp, eta), (Mchirp, q)
# compute the other two if not yet present
# p can be a posterior or a single set of source params
def complete_mass_params(p_in):
    p = dict(p_in)
    if ('m1' in p.keys() and 'm2' in p.keys()):
        if not 'Mchirp' in p.keys():
            p['Mchirp'] = Mchirpofm1m2(p['m1'], p['m2'])
        if not 'eta' in p.keys():
            p['eta'] = etaofm1m2(p['m1'], p['m2'])
        if not 'M' in p.keys():
            p['M'] = Mofm1m2(p['m1'], p['m2'])
        if not 'q' in p.keys():
            p['q'] = qofm1m2(p['m1'], p['m2'])
    elif ('Mchirp' in p.keys() and 'eta' in p.keys()):
        if not 'm1' in p.keys():
            p['m1'] = m1ofMchirpeta(p['Mchirp'], p['eta'])
        if not 'm2' in p.keys():
            p['m2'] = m2ofMchirpeta(p['Mchirp'], p['eta'])
        if not 'M' in p.keys():
            p['M'] = MofMchirpeta(p['Mchirp'], p['eta'])
        if not 'q' in p.keys():
            p['q'] = qofMchirpeta(p['Mchirp'], p['eta'])
    elif ('M' in p.keys() and 'q' in p.keys()):
        if not 'm1' in p.keys():
            p['m1'] = m1ofMq(p['M'], p['q'])
        if not 'm2' in p.keys():
            p['m2'] = m2ofMq(p['M'], p['q'])
        if not 'Mchirp' in p.keys():
            p['Mchirp'] = MchirpofMq(p['M'], p['q'])
        if not 'eta' in p.keys():
            p['eta'] = etaofMq(p['M'], p['q'])
    elif ('Mchirp' in p.keys() and 'q' in p.keys()):
        eta = etaofq(p['q'])
        if not 'm1' in p.keys():
            p['m1'] = m1ofMchirpeta(p['Mchirp'], eta)
        if not 'm2' in p.keys():
            p['m2'] = m2ofMchirpeta(p['Mchirp'], eta)
        if not 'M' in p.keys():
            p['M'] = MofMchirpeta(p['Mchirp'], eta)
        if not 'eta' in p.keys():
            p['eta'] = eta
    else:
        raise ValueError('Incomplete mass parameters.')
    return p
# Complete spin parameters: from any set (chi1, chi2), (chis, chia), (chip, chim)
# compute the other two if not yet present
# p can be a posterior or a single set of source params
# Assumes input has mass parameters for chip, chim
def complete_spin_params(p_in):
    p = dict(p_in)
    if ('chi1' in p.keys() and 'chi2' in p.keys()):
        if not ('chis' in p.keys() and 'chia' in p.keys()):
            p['chis'] = chisofchi1chi2(p['chi1'], p['chi2'])
            p['chia'] = chiaofchi1chi2(p['chi1'], p['chi2'])
        if not ('chip' in p.keys() and 'chim' in p.keys()):
            p['chip'] = chipofchi1chi2(p['chi1'], p['chi2'], p['q'])
            p['chim'] = chimofchi1chi2(p['chi1'], p['chi2'], p['q'])
        if not 'chiPN' in p.keys():
            p['chiPN'] = chiPNofchi1chi2(p['chi1'], p['chi2'], p['q'])
    elif ('chis' in p.keys() and 'chia' in p.keys()):
        if not ('chi1' in p.keys() and 'chi2' in p.keys()):
            p['chi1'] = chi1ofchischia(p['chis'], p['chia'])
            p['chi2'] = chi2ofchischia(p['chis'], p['chia'])
        if not ('chip' in p.keys() and 'chim' in p.keys()):
            p['chip'] = chipofchischia(p['chis'], p['chia'], p['q'])
            p['chim'] = chimofchischia(p['chis'], p['chia'], p['q'])
        if not 'chiPN' in p.keys():
            p['chiPN'] = chiPNofchischia(p['chis'], p['chia'], p['q'])
    elif ('chip' in p.keys() and 'chim' in p.keys()):
        if not ('chi1' in p.keys() and 'chi2' in p.keys()):
            p['chi1'] = chi1ofchipchim(p['chip'], p['chim'], p['q'])
            p['chi2'] = chi2ofchipchim(p['chip'], p['chim'], p['q'])
        if not ('chis' in p.keys() and 'chia' in p.keys()):
            p['chis'] = chisofchipchim(p['chip'], p['chim'], p['q'])
            p['chia'] = chiaofchipchim(p['chip'], p['chim'], p['q'])
        if not 'chiPN' in p.keys():
            p['chiPN'] = chiPNofchipchim(p['chip'], p['chim'], p['q'])
    elif ('chiPN' in p.keys() and 'chim' in p.keys()):
        p['chip'] = chipofchiPNchim(p['chiPN'], p['chim'], p['q'])
        if not ('chi1' in p.keys() and 'chi2' in p.keys()):
            p['chi1'] = chi1ofchipchim(p['chip'], p['chim'], p['q'])
            p['chi2'] = chi2ofchipchim(p['chip'], p['chim'], p['q'])
        if not ('chis' in p.keys() and 'chia' in p.keys()):
            p['chis'] = chisofchipchim(p['chip'], p['chim'], p['q'])
            p['chia'] = chiaofchipchim(p['chip'], p['chim'], p['q'])
    else:
        raise ValueError('Incomplete spin parameters.')
    return p
# Complete AGN params from the set of params used in PE
def complete_AGN_params(p_in):
    p = dict(p_in)
    R_AGN_s = np.sqrt(p['Rcostheta_AGN']**2 + p['Rsintheta_AGN']**2)
    M_AGN_s = p['Omega_AGN']**2 * R_AGN_s**3
    p['M_AGN'] = M_AGN_s / pyconstants.MTSUN_SI
    p['R_AGN'] = R_AGN_s / M_AGN_s
    p['theta_AGN'] = np.arctan2(p['Rsintheta_AGN'], p['Rcostheta_AGN'])
    return p

# Newtonian estimate of the relation Mf(deltat/M) (for the 22 mode) - gives the starting geometric frequency for a given mass ratio and a given geometric duration of the observations
def funcNewtonianfoftGeom(q, deltat):
    nu = q/(1. + q)/(1. + q)
    return 1./np.pi*(256*nu/5.*deltat)**(-3./8)
# Newtonian estimate of the relation f(deltat) (for the 22 mode) - gives the starting geometric frequency for a given mass ratio and a given geometric duration of the observations-output in Hz -- m1,m2,t in solar masses and s
def funcNewtonianfoft(m1, m2, deltat):
    mtot = m1 + m2
    q = m1/m2
    return funcNewtonianfoftGeom(q, deltat/(mtot * pyconstants.MTSUN_SI))/(mtot * pyconstants.MTSUN_SI)
# Newtonian estimate of the relation deltat/M(Mf) (for the 22 mode) - gives the time to merger from a given starting geometric frequency for a given mass ratio
def funcNewtoniantoffGeom(q, f):
    nu = q/(1. + q)/(1. + q)
    return 5./256/nu*(np.pi*f)**(-8./3)
# Newtonian estimate of the relation deltat(f) (for the 22 mode) - gives the time to merger from a given starting  frequency for a given mass ratio - output in s -- m1,m2,f in solar masses and Hz
def funcNewtoniantoff(m1, m2, f):
    mtot = m1 + m2
    q = m1/m2
    return funcNewtoniantoffGeom(q, (mtot * pyconstants.MTSUN_SI)*f)*(mtot * pyconstants.MTSUN_SI)

################################################################################
# Various utilities
################################################################################

# Logarithmic sampling
def logspace(start, stop, nb):
    ratio = (stop/start)**(1./(nb-1))
    vals = start * np.power(ratio, np.arange(nb))
    # Eliminate rounding errors at extremities
    vals[0] = start
    vals[-1] = stop
    return vals
# Sqrt sampling
def sqrtspace(start, stop, nb):
    sqrtvals = np.linspace(np.sqrt(start), np.sqrt(stop), nb)
    vals = np.power(sqrtvals, 2)
    # Eliminate rounding errors at extremities
    vals[0] = start
    vals[-1] = stop
    return vals

# mod functions for angles
# Mod 2pi - shifted to return a result in [-pi, pi[
def mod2pi(x):
    rem = np.remainder(x, 2*np.pi)
    if isinstance(x, np.ndarray) and x.shape is not ():
        mask = (rem>np.pi)
        rem[mask] -= 2*np.pi
    else:
        if rem>np.pi:
            rem -= 2*np.pi
    return rem
# Mod pi
def modpi(x):
    return np.remainder(x, np.pi)
# Mod pi, result in ]-pi/2, pi/2]
def modpi2(x):
    return np.pi/2 - np.remainder(np.pi/2 - x, np.pi)
# Mod with generic modulus, result symmetric around 0
def modsym(x, a):
    rem = np.remainder(x, a)
    if isinstance(x, np.ndarray) and x.shape is not ():
        mask = (rem>a/2.)
        rem[mask] -= a
    else:
        if rem>np.pi:
            rem -= a
    return rem
def mod_interval(x, interval=None):
    """
    Remainder of x (can be a vector) on an interval ]a,b]
    Args:
      x          # Input data, can be a vector (no check)
    Kwargs:
      interval   # Target interval, list [a,b] for mod in ]a,b]
                   (default None, ignore)
    """
    if interval is None:
        return x
    else:
        a, b = interval
        return b - np.remainder(b-x, b-a)

# Fold a list : get n elements by cycling through the list
def fold_list(xlist, n):
    l = len(xlist)
    return (xlist * (n//l + 1))[:n]

# Restrict data to an interval according to first column
# If data[:,0]=x and interval=[a,b], selects data such that a<=x<=b
# Typically results in data being entirely inside interval
def restrict_data(data, interval):
    if not interval: # test if interval==[]
        return data
    else:
        # Make an initial guess based on global length - then adjust starting and ending indices
        x = data[:,0]
        n = len(data)
        deltax = (x[-1] - x[0]) / n
        if interval[0] < x[0]:
            ibeg = 0
        else:
            ibeg = min(int((interval[0]-x[0]) / deltax), n-1)
            while ibeg < n and x[ibeg] < interval[0]:
                ibeg += 1
            while ibeg > 0 and x[ibeg-1] > interval[0]:
                ibeg -= 1
        if interval[-1] > x[-1]:
            iend = n-1
        else:
            iend = n-1 - min(int((x[-1] - interval[-1]) / deltax), n-1)
            while iend > 0 and x[iend] > interval[-1]:
                iend -= 1
            while iend < n-1 and x[iend+1] < interval[-1]:
                iend += 1
        return data[ibeg:iend+1]
# Restrict data to an interval according to first column
# If data[:,0]=x and interval=[a,b], selects data from the last point such that x<=a to the first point such that b<=x
# Typically results in interval being entirely covered by data
def restrict_data_soft(data, interval):
    if not interval: # test if interval==[]
        return data
    else:
        # Make an initial guess based on global length - then adjust starting and ending indices
        x = data[:,0]
        n = len(data)
        deltax = (x[-1] - x[0]) / n
        if interval[0] < x[0]:
            ibeg = 0
        else:
            ibeg = min(int((interval[0]-x[0]) / deltax), n-1)
            while ibeg < n-1 and x[ibeg+1] <= interval[0]:
                ibeg += 1
            while ibeg > 0 and x[ibeg] > interval[0]:
                ibeg -= 1
        if interval[-1] > x[-1]:
            iend = n-1
        else:
            iend = n-1 - min(int((x[-1] - interval[-1]) / deltax), n-1)
            while iend > 0 and x[iend-1] >= interval[-1]:
                iend -= 1
            while iend < n-1 and x[iend+1] < interval[-1]:
                iend += 1
        return data[ibeg:iend+1]
# Trim zeros from a waveform according to chosen columns
def trim_zeros_bycol(data, cols, ifallzero_returnvoid=False):
    ilastnonzero = len(data)-1
    while (ilastnonzero>=0) and all([data[ilastnonzero, i]==0. for i in cols]):
        ilastnonzero -= 1
    ifirstnonzero = 0
    while (ifirstnonzero<=len(data)-1) and all([data[ifirstnonzero, i]==0. for i in cols]):
        ifirstnonzero += 1
    if ifirstnonzero>ilastnonzero and not ifallzero_returnvoid: # if everything is zero, do nothing
        return data
    else:
        return data[ifirstnonzero:ilastnonzero+1]
# Restrict data according to given column
# if x=data[:,col], return data such that interval[0]<=x<=interval[1]
def restrict_data_bycol(data, interval, col=0):
    if not interval: # test if interval==[]
        return data
    else:
        x = data[:,col]
        mask = np.logical_and(interval[0]<=x, x<=interval[1])
        return data[mask]

################################################################################
# Utilities for FFT/IFFT
################################################################################

# Definitions for the windowing function -- function for vectorial argument
# In order to avoid overflows in the exponentials, we set the boundaries (di, df) so that anything below 10^-20 is considered zero
def window_planck_vec(x, xi, xf, deltaxi, deltaxf):
    di = deltaxi/(20*np.log(10))
    df = deltaxf/(20*np.log(10))
    w = np.zeros(len(x), dtype=float)
    #
    mask = np.logical_or(x <= xi + di, x >= xf - df)
    w[mask] = 0.
    #
    mask = np.logical_and(xi + di < x, x < xi + deltaxi - di)
    xm = x[mask]
    w[mask] = 1./(1 + np.exp(deltaxi/(xm - xi) + deltaxi/(xm - (xi + deltaxi))))
    #
    mask = np.logical_and(xi + deltaxi - di <= x, x <= xf - deltaxf + df)
    w[mask] = 1.
    #
    mask = np.logical_and(xf - deltaxf + df < x, x < xf - df)
    xm = x[mask]
    w[mask] = 1./(1 + np.exp(-(deltaxf/(xm - (xf - deltaxf))) - deltaxf/(xm - xf)))
    #
    return w

# The FFT function - accepts real or complex (real+real) input, discards negative frequencies
# Fourier convention: Ftilde(f) = \int dt exp(+2 i pi f t) * F(t)
def fft_positivef(timeseries):
    n = len(timeseries)
    ncol = timeseries.shape[1] - 1
    deltat = timeseries[1,0] - timeseries[0,0]
    deltaf = 1./(n*deltat)
    # Fast Fourier Transform
    freqs = deltaf*np.arange(n)
    # Input values for the fft - accomodate for the real and complex cases
    if ncol==1:
        vals = timeseries[:,1]
    elif ncol==2:
        vals = timeseries[:,1] + 1j*timeseries[:,2]
    else:
        raise Exception('Incorrect number of columns in array.')
    # BEWARE: due to the different convention for the sign of Fourier frequencies, we have to reverse the FFT output
    # Beware also that the FFT treats effectively the initial time as 0
    # BEWARE: in the reversion of the numpy-convention FFT output, we have to set aside the 0-frequency term
    fftvals_np = deltat * np.fft.fft(vals)
    fdvals = np.zeros_like(fftvals_np, dtype=complex)
    fdvals[0] = fftvals_np[0]
    fdvals[1:] = fftvals_np[1:][::-1]
    # Discarding information on negative frequencies - if real timeseries in input, no information loss as the deleted values are the conjugate
    fdvals = fdvals[:n//2]
    freqs = freqs[:n//2]
    # Coming back to the initial times
    tshift = timeseries[0,0]
    fac_timeshift = np.exp(1j*2*np.pi*freqs*tshift)
    fdvals = fdvals * fac_timeshift
    freqseries = np.array([freqs, np.real(fdvals), np.imag(fdvals)]).T
    return freqseries
# The IFFT function - input format (f, real, imag) expected to be FFT of complex data, assumes negative frequencies are zero (not conjugated as for the FFT of real data)
# Fourier convention: Ftilde(f) = \int dt exp(+2 i pi f t) * F(t)
# WORK IN PROGRESS
def ifft_positivef(freqseries):
    n = len(freqseries)
    ncol = freqseries.shape[1] - 1
    deltaf = freqseries[1,0] - freqseries[0,0]
    # Input values for the ifft - has no reason to bea real, check we have real and imag
    if not ncol==2:
        raise Exception('Incorrect number of columns in array.')
    # Zero-pad to next power of 2, then 0-pad by factor 2 for the negative frequencies
    freqseries_pad = zeropad(freqseries, extend=1)
    npad = len(freqseries_pad)
    # 1D FD values
    fdvals = freqseries_pad[:,1] + 1j*freqseries_pad[:,2]
    # BEWARE: due to the different convention for the sign of Fourier frequencies, we have to reverse the IFFT input
    # Beware also that the FFT treats effectively the initial time as 0
    # BEWARE: in the reversion of the numpy-convention IFFT input, we have to set aside the 0-frequency term
    fdvals_np = np.zeros_like(fdvals, dtype=complex)
    fdvals_np[0] = fdvals[0]
    fdvals_np[1:] = fdvals[1:][::-1]
    # Inverse FFT
    deltat = 1./(npad*deltaf)
    ifftvals_np = 1./deltat * np.fft.ifft(fdvals_np)
    # Rebuild time series from positive and negative times
    tdvals = np.concatenate((ifftvals_np[npad//2:], ifftvals_np[:npad//2]))
    # Rebuild times
    times = deltat*np.arange(-npad//2, npad//2)
    timeseries = np.array([times, np.real(tdvals), np.imag(tdvals)]).T
    return timeseries
# The IFFT function - input format (f, real, imag) expected to be FFT of real data, assumes negative frequencies are conjugate (as for the FFT of real data)
# Fourier convention: Ftilde(f) = \int dt exp(+2 i pi f t) * F(t)
# WORK IN PROGRESS
# NOTE: could be implemented with irfft, here for simplicity we keep close to the already-tested setting of the IFFT for complex TD data
def ifft_real(freqseries):
    n = len(freqseries)
    ncol = freqseries.shape[1] - 1
    deltaf = freqseries[1,0] - freqseries[0,0]
    # Input values for the ifft - has no reason to bea real, check we have real and imag
    if not ncol==2:
        raise Exception('Incorrect number of columns in array.')
    # Zero-pad to next power of 2, 0-pad by factor 2 to make room for the negative frequencies (conjugate data)
    freqseries_pad = zeropad(freqseries, extend=1)
    npad = len(freqseries_pad)
    # 1D FD values
    fdvals = freqseries_pad[:,1] + 1j*freqseries_pad[:,2]
    # Restore negative frequencies as conjugate of values for positive frequencies
    # NOTE: both values at 0 and n//2 are real and untouched here
    fdvals[npad//2+1:] = np.conjugate(fdvals[1:npad//2][::-1])
    # BEWARE: due to the different convention for the sign of Fourier frequencies, we have to reverse the IFFT input
    # Beware also that the FFT treats effectively the initial time as 0
    # BEWARE: in the reversion of the numpy-convention IFFT input, we have to set aside the 0-frequency term
    fdvals_np = np.zeros_like(fdvals, dtype=complex)
    fdvals_np[0] = fdvals[0]
    fdvals_np[1:] = fdvals[1:][::-1]
    # Inverse FFT
    deltat = 1./(npad*deltaf)
    ifftvals_np = 1./deltat * np.fft.ifft(fdvals_np)
    # Rebuild time series from positive and negative times
    tdvals = np.concatenate((ifftvals_np[npad//2:], ifftvals_np[:npad//2]))
    # Rebuild times
    times = deltat*np.arange(-npad//2, npad//2)
    # Discard imaginary part, must be 0 to machine precision
    timeseries_real = np.array([times, np.real(tdvals)]).T
    return timeseries_real

# Function for zero-padding a real array at the end
# Assumes a constant x-spacing - works with any number of columns
def zeropad(series, extend=0):
    n = len(series)
    ncol = series.shape[1]
    deltax = series[1,0] - series[0,0]
    xf = series[-1,0]
    nzeros = pow(2, extend + int(np.ceil(np.log(n)/np.log(2)))) - n
    res = np.zeros((n+nzeros, ncol), dtype=float)
    xextension = xf + deltax*np.arange(1, nzeros+1)
    res[:n, :] = series
    res[n:, 0] = xextension
    return res

################################################################################
# Stationary Phase Approximation (SPA)
################################################################################

# Stationary Phase Approximation for a time-domain signal
# Allows for chirping or anti-chirping signal with the sign option
def signal_spa(signal_td, sign=1):
    t, amp, phi = signal_td.T
    spline_phi = pyspline.CubicSpline(t, phi)
    # NOTE: here omega means phidot, not omega_orb
    omega = spline_phi.get_spline_d()[:,1]
    omegadot = spline_phi.get_spline_dd()[:,1]
    f = -omega / (2*np.pi)
    A_spa = amp * np.sqrt(2*np.pi / (-sign*omegadot))
    Psi_spa = -phi - 2*np.pi*f*t - sign*np.pi/4
    signal_fd = np.array([f, A_spa, -Psi_spa]).T
    if sign==1:
        return signal_fd
    else:
        signal_fd_reverse = np.zeros_like(signal_fd)
        signal_fd_reverse[:] = signal_fd[::-1]
        return signal_fd_reverse

# Inverse Stationary Phase Approximation for a Fourier-domain signal
# Allows for chirping or anti-chirping signal with the sign option
def signal_inverse_spa(signal_fd, sign=1):
    f, amp_fd, phi_fd = signal_fd.T
    spline_phi_fd = pyspline.CubicSpline(f, phi_fd)
    tf = 1./(2*np.pi) * spline_phi_fd.get_spline_d()[:,1]
    # NOTE: Tf^2 here is positive by convention
    Tf2 = sign * 1./(2*np.pi)**2 * spline_phi_fd.get_spline_dd()[:,1]
    # NOTE: here omega means phidot, not omega_orb
    omegadot_abs = 1./Tf2
    amp_td = amp_fd * np.sqrt(omegadot_abs / (2*np.pi))
    # NOTE: phi_fd = -Psi
    phi_td = phi_fd - 2*np.pi*f*tf - sign*np.pi/4
    signal_td = np.array([tf, amp_td, phi_td]).T
    return signal_td

################################################################################
# Utilities for frequency scaling between mode when not using m/2*f22
################################################################################

# Scales an array of frequencies to new bounds, using log-affine scaling
def log_affine_scaling(freq, fLow, fHigh):
    lp1 = np.log(fLow)
    lp2 = np.log(fHigh)
    l = np.log(freq)
    l1 = l[0]
    l2 = l[-1]
    a = (lp2 - lp1) / (l2 - l1)
    b = (lp1*l2 - l1*lp2) / (l2 - l1)
    lp = a*l + b
    return np.exp(lp)

################################################################################
# Utilities for decomposing a signal in chirping and anti-chirping bands
################################################################################

def func_sign_interval(y, interval):
    if y < interval[0]: return -1
    elif interval[0] <= y <= interval[1]: return 0
    else: return 1

def func_solve_linear_interp(X1, X2, yval):
    x1, y1 = X1
    x2, y2 = X2
    xval = x1 + (x2 - x1) / (y2 - y1) * (yval - y1)
    return xval

def func_domain_decomp_linear(x, y, interval):
    N = len(x)
    nband_max = 2*N - 1
    iband = 0
    bands = np.zeros((nband_max,3), dtype=float)
    sign = func_sign_interval(y[0], interval)
    bands[0][0] = x[0]
    bands[0][2] = sign
    for i in range(1,N):
        sign_new = func_sign_interval(y[i], interval)
        sign_diff = sign_new - sign
        X1 = [x[i-1], y[i-1]]
        X2 = [x[i], y[i]]
        if np.abs(sign_diff)==1:
            if sign==-1 or sign_new==-1:
                yval = interval[0]
            elif sign==1 or sign_new==1:
                yval = interval[1]
            xval = func_solve_linear_interp(X1, X2, yval)
            bands[iband][1] = xval
            bands[iband+1][0] = xval
            bands[iband+1][2] = sign_new
            iband += 1
        elif np.abs(sign_diff)==2:
            if sign==-1:
                yval1, yval2 = interval[0], interval[1]
            elif sign==1:
                yval1, yval2 = interval[1], interval[0]
            xval1 = func_solve_linear_interp(X1, X2, yval1)
            xval2 = func_solve_linear_interp(X1, X2, yval2)
            bands[iband][1] = xval1
            bands[iband+1][0] = xval1
            bands[iband+1][1] = xval2
            bands[iband+1][2] = 0
            bands[iband+2][0] = xval2
            bands[iband+2][2] = sign_new
            iband += 2
        else:
            pass
        sign = sign_new
    bands[iband][1] = x[-1]
    bands = bands[:iband+1]
    return bands

################################################################################
# I/O utilities
################################################################################

# Write a dictionary of type {str:val} to hdf5 as separate 1-element datasets
# At first sight, no way around using 1-element numpy arrays...
# TODO: support string types
def write_h5py_dict(h5gr, dic):
    for p in dic:
        h5gr.create_dataset(p, data=np.array([dic[p]]))
    return
def read_h5py_dict(h5gr):
    dic = {}
    for p in list(h5gr.keys()):
        dic[p] = h5gr[p][:][0] # We assume datasets are 1-element arrays
    return dic

################################################################################
# Spin-weighted spherical harmonics
################################################################################

cpdef sYlm(l, m, theta, phi, s=-2):
    return SpinWeightedSphericalHarmonic (
        theta,
        phi,
        s,
        l,
        m
    )

################################################################################
# Functions to build frequency grids
################################################################################

cdef class WaveformGeomFrequencyGrid:
    """ Geom frequency grid suitable for waveform amp/phase interpolation.
    """

    cdef real_vector* CMfreq  # pointer to C frequency vector

    cdef Mfmin                # Minimal geometric frequency
    cdef Mfmax                # Maximal geometric frequency
    cdef eta                  # Symmetric mass ratio
    cdef length               # Size of the frequency data
    cdef acc                  # Target accuracy for inspiral

    cdef Mfreq                # numpy array of frequencies


    def __init__(self, Mfmin, Mfmax, eta, acc=1e-4):
        """Constructor
        Args:
          Mfmin                 # Minimal geometric frequency
          Mfmax                 # Maximal geometric frequency
          eta                   # Symmetric mass ratio
        Keyword args:
          acc                   # Target accuracy for inspiral (default 1e-4)
        """
        self.Mfmin = Mfmin
        self.Mfmax = Mfmax
        self.eta = eta
        self.acc = acc

        self.CMfreq = NULL

        ret = BuildWaveformGeomFrequencyGrid(
            &self.CMfreq,
            self.Mfmin,
            self.Mfmax,
            self.eta,
            self.acc
        );
        if ret == _FAILURE:
            raise ValueError("Call to BuildWaveformGeomFrequencyGrid() failed.")

        # Cast C double array to numpy via a MemoryView
        cdef double[::1] view_Mfreq = \
          <(double)[:self.CMfreq.size]> self.CMfreq.data
        self.Mfreq = np.asarray(view_Mfreq)
        #self.Mfreq = real_vector_to_np_array(self.CMfreq)

    def __dealloc__(self):
        """Destructor
        """
        if self.CMfreq != NULL:
            real_vector_free(self.CMfreq)

    def get_Mfreq(self):
        return np.copy(self.Mfreq)

cdef class ResponseFrequencyGrid:
    """ Geom frequency grid suitable for waveform amp/phase interpolation.
    """

    cdef real_vector* Cfreq     # pointer to C frequency vector
    cdef real_matrix* Ctfspline # pointer to C spline matrix for tf

    cdef f_min                  # Minimal frequency (default None: use tf)
    cdef f_max                  # Maximal frequency (default None: use tf)
    cdef Deltat_max             # Maximal time step, in years (default 2 weeks)
    cdef Deltaf_max             # Maximal frequency step, in Hz (default 1mHz)
    cdef nptlogmin              # Minimal number of points with log-spacing

    cdef freq                   # numpy array of frequencies

    def __init__(self, np.ndarray[np.float_t, ndim=2] tfspline,
                       f_min=None, f_max=None, Deltat_max=0.02083335,
                       Deltaf_max=0.001, nptlogmin=150):
        """Constructor
        Args:
          tfspline              # Input tf cubic spline (numpy array)
        Keyword args:
          f_min                 # Minimal frequency (default None: use tf)
          f_max                 # Maximal frequency (default None: use tf)
          Deltat_max            # Maximal time step, in years (default 2 weeks)
          Deltaf_max            # Maximal frequency step, in Hz (default 1mHz)
          nptlogmin             # Minimal number of points with log-spacing
        """
        n_tf = tfspline.shape[0]
        f_min_tf = tfspline[0, 0]
        f_max_tf = tfspline[n_tf-1, 0]
        if f_min is None:
            self.f_min = f_min_tf
        else:
            self.f_min = f_min
        if f_max is None:
            self.f_max = f_max_tf
        else:
            self.f_max = f_max
        self.Deltat_max = Deltat_max
        self.Deltaf_max = Deltaf_max
        self.nptlogmin = nptlogmin

        self.Cfreq = NULL

        # Build a real_matrix representation of the input numpy array spline
        if not tfspline.flags['C_CONTIGUOUS']:
            raise ValueError('Input np array tfspline is not C_CONTIGUOUS.')
        cdef double* tfspline_data = <double *> &tfspline[0,0]
        self.Ctfspline = real_matrix_view(tfspline_data,
                                      tfspline.shape[0], tfspline.shape[1])

        ret = BuildResponseFrequencyGrid(
            &self.Cfreq,
            self.Ctfspline,
            self.f_min,
            self.f_max,
            self.Deltat_max,
            self.Deltaf_max,
            self.nptlogmin
        );
        if ret == _FAILURE:
            raise ValueError("Call to BuildResponseFrequencyGrid() failed.")

        # Cast C double array to numpy via a MemoryView
        cdef double[::1] view_freq = \
          <(double)[:self.Cfreq.size]> self.Cfreq.data
        self.freq = np.asarray(view_freq)
        #self.freq = real_vector_to_np_array(self.Cfreq)

    def __dealloc__(self):
        """Destructor
        """
        if self.Cfreq != NULL:
            real_vector_free(self.Cfreq)

    def get_freq(self):
        return np.copy(self.freq)

cdef class FrequencyGrid:
    """ Frequency grid suitable for both the waveform and the response.
    """

    cdef real_vector* Cfreq     # pointer to C frequency vector

    cdef f_min                  # Minimal frequency (default None: use tf)
    cdef f_max                  # Maximal frequency (default None: use tf)
    cdef M                      # Total mass (solar masses)
    cdef q                      # Mass ratio
    cdef Deltat_max             # Maximal time step, in years (default 2 weeks)
    cdef Deltaf_max             # Maximal frequency step, in Hz (default 1mHz)
    cdef DeltalnMf_max          # Maximal ln(Mf) step (default 0.025)
    cdef acc                    # Target phase interp. error (default 1e-4)
    cdef nptlogmin              # Minimal number of points (default 100)

    cdef freq                   # numpy array of frequencies

    def __init__(self, f_min, f_max, M, q,
                       Deltat_max=0.02083335, Deltaf_max=0.001,
                       DeltalnMf_max=0.025,
                       acc=1e-4, nptlogmin=100):
        """Constructor
        Args:
          f_min                 # Minimal frequency (Hz)
          f_max                 # Maximal frequency (Hz)
          M                     # Total mass (solar masses)
          q                     # Mass ratio
        Keyword args:
          Deltat_max            # Maximal time step, in years (default 2 weeks)
          Deltaf_max            # Maximal frequency step, in Hz (default 1mHz)
          DeltalnMf_max         # Maximal ln(Mf) step (default 0.025)
          acc                   # Target phase interp. error (default 1e-4)
          nptlogmin             # Minimal number of points (default 100)
        """
        self.f_min = f_min
        self.f_max = f_max
        self.M = M
        self.q = q
        self.Deltat_max = Deltat_max
        self.Deltaf_max = Deltaf_max
        self.DeltalnMf_max = DeltalnMf_max
        self.acc = acc
        self.nptlogmin = nptlogmin

        self.Cfreq = NULL

        ret = BuildFrequencyGrid(
            &self.Cfreq,
            self.f_min,
            self.f_max,
            self.M,
            self.q,
            self.Deltat_max,
            self.Deltaf_max,
            self.DeltalnMf_max,
            self.acc,
            self.nptlogmin
        );
        if ret == _FAILURE:
            raise ValueError("Call to BuildFrequencyGrid() failed.")

        # Cast C double array to numpy via a MemoryView
        cdef double[::1] view_freq = \
          <(double)[:self.Cfreq.size]> self.Cfreq.data
        self.freq = np.asarray(view_freq)
        #self.freq = real_vector_to_np_array(self.Cfreq)

    def __dealloc__(self):
        """Destructor
        """
        if self.Cfreq != NULL:
            real_vector_free(self.Cfreq)

    def get_freq(self):
        return np.copy(self.freq)

cdef class MergeGrids:
    """ Merge two grids with increasing values (e.g. two frequency grids)
    Discards duplicate values
    Allows for safeguarding: can ignore points that are almost identical
    """

    cdef real_vector* Cgrid   # pointer to output C vector
    cdef real_vector* Cgrid1  # pointer to inupt C vector 1
    cdef real_vector* Cgrid2  # pointer to input C vector 2

    cdef xmin                 # Minimal value
    cdef xmax                 # Maximal value
    cdef usedeltaxmin         # Use minimal x step
    cdef usedeltalnxmin       # Use minimal x relative step in ln
    cdef deltaxmin            # Minimal x step
    cdef deltalnxmin          # Minimal x relative step in ln

    cdef grid                 # numpy array of frequencies


    def __init__(self,
                 np.ndarray[np.float_t, ndim=1] grid1,
                 np.ndarray[np.float_t, ndim=1] grid2,
                 deltaxmin=None, deltalnxmin=-11.512925465):
        """Constructor
        Args:
          grid1                # First vector of frequencies to be merged
          grid2                # Second vector of frequencies to be merged
        Keyword args:
          xmin                  # Minimal value
          xmax                  # Maximal value
          deltaxmin             # Minimal step
          deltalnxmin           # Minimal step ratio, in ln (default ratio=1e-5)
        """
        if deltaxmin is not None:
            self.usedeltaxmin = True
            self.deltaxmin = deltaxmin
        else:
            self.usedeltaxmin = False
            self.deltaxmin = 0. # will be ignored
        if deltalnxmin is not None:
            self.usedeltalnxmin = True
            self.deltalnxmin = deltalnxmin
        else:
            self.usedeltalnxmin = False
            self.deltalnxmin = 0. # will be ignored
        cdef int Cusedeltaxmin = <int> self.usedeltaxmin
        cdef int Cusedeltalnxmin = <int> self.usedeltalnxmin

        if not grid1.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array grid1 is not C_CONTIGUOUS')
        if not grid2.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array grid2 is not C_CONTIGUOUS')
        cdef double* grid1_data = <double *> &grid1[0]
        self.Cgrid1 = real_vector_view(grid1_data, grid1.shape[0])
        cdef double* grid2_data = <double *> &grid2[0]
        self.Cgrid2 = real_vector_view(grid2_data, grid2.shape[0])

        self.Cgrid = NULL

        ret = BuildMergedGrid(
            &self.Cgrid,
            self.Cgrid1,
            self.Cgrid2,
            Cusedeltaxmin,
            self.deltaxmin,
            Cusedeltalnxmin,
            self.deltalnxmin
        );
        if ret == _FAILURE:
            raise ValueError("Call to BuildMergedGrid() failed.")

        # Cast C double array to numpy via a MemoryView
        cdef double[::1] view_grid = \
          <(double)[:self.Cgrid.size]> self.Cgrid.data
        self.grid = np.asarray(view_grid)

    def __dealloc__(self):
        """Destructor
        """
        if self.Cgrid != NULL:
            real_vector_free(self.Cgrid)

    def get_grid(self):
        return np.copy(self.grid)

# Function that uses the 1/(2T) Nyquist sampling, then breaks to log-sampling
# No C extension, used primarily for Fisher where it is not the dominating cost
def composite_nyquist_log_freq_grid(f0, f1, m1, m2, Deltalnf=5e-5, refine=0):
    Ms = (m1 + m2) * pyconstants.MTSUN_SI
    nu = etaofm1m2(m1, m2)
    lambd = 128*nu/3. * 1./Ms * (np.pi*Ms)**(8./3)
    xi = np.exp(Deltalnf) - 1.
    f_break = (3*lambd/5./xi)**(-3./5)
    if f_break > f1:
        N_nyq = int(np.floor(1./lambd * (f0**(-5./3) - f1**(-5./3)))) + 1
        freq = np.zeros(N_nyq, dtype=float)
        n = np.arange(N_nyq-1)
        freq[:N_nyq-1] = (f0**(-5./3) - lambd*n)**(-3./5)
        freq[N_nyq-1] = f1
    elif f_break > f0:
        N_nyq = int(np.floor(1./lambd * (f0**(-5./3) - f_break**(-5./3)))) # Does not include f_break
        N_log = int(np.ceil(np.log(f1/f_break) / Deltalnf))
        freq = np.zeros(N_nyq + N_log, dtype=float)
        n = np.arange(N_nyq)
        freq[:N_nyq] = (f0**(-5./3) - lambd*n)**(-3./5)
        freq[N_nyq:] = np.geomspace(f_break, f1, N_log)
    else:
        N_log = int(np.ceil(np.log(f1/f0) / Deltalnf))
        freq = np.geomspace(f0, f1, N_log)
    while refine > 0:
        freq_refine = np.zeros(2*len(freq) - 1, dtype=float)
        freq_refine[0::2] = freq
        freq_refine[1::2] = (freq[1:] + freq[:-1]) / 2.
        freq = freq_refine
        refine -= 1
    return freq

################################################################################
# Helper function for residuals likelihood
################################################################################

cpdef linear_interp_multimode_3chan_resid_norm(
        np.ndarray[np.complex_t, ndim=3] alpha0,
        np.ndarray[np.complex_t, ndim=3] alpha1,
        np.ndarray[np.complex_t, ndim=3] w0,
        np.ndarray[np.complex_t, ndim=3] w1,
        np.ndarray[np.complex_t, ndim=3] w2):
    """ Helper function for residuals likelihood.
    """

    cdef complex_array_3d* Calpha0  # pointer to C alpha0
    cdef complex_array_3d* Calpha1  # pointer to C alpha1
    cdef complex_array_3d* Cw0      # pointer to C w0
    cdef complex_array_3d* Cw1      # pointer to C w1
    cdef complex_array_3d* Cw2      # pointer to C w2

    # Build a complex_array_3d representation of the input numpy arrays
    if not alpha0.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array alpha0 is not C_CONTIGUOUS')
    if not alpha1.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array alpha1 is not C_CONTIGUOUS')
    if not w0.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array w0 is not C_CONTIGUOUS')
    if not w1.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array w1 is not C_CONTIGUOUS')
    if not w2.flags['C_CONTIGUOUS']:
        raise ValueError('Input numpy array w2 is not C_CONTIGUOUS')
    cdef double complex* alpha0_data = <double complex *> &alpha0[0,0,0]
    Calpha0 = complex_array_3d_view(alpha0_data, alpha0.shape[0],
                                              alpha0.shape[1], alpha0.shape[2])
    cdef double complex* alpha1_data = <double complex *> &alpha1[0,0,0]
    Calpha1 = complex_array_3d_view(alpha1_data, alpha1.shape[0],
                                              alpha1.shape[1], alpha1.shape[2])
    cdef double complex* w0_data = <double complex *> &w0[0,0,0]
    Cw0 = complex_array_3d_view(w0_data, w0.shape[0], w0.shape[1], w0.shape[2])
    cdef double complex* w1_data = <double complex *> &w1[0,0,0]
    Cw1 = complex_array_3d_view(w1_data, w1.shape[0], w1.shape[1], w1.shape[2])
    cdef double complex* w2_data = <double complex *> &w2[0,0,0]
    Cw2 = complex_array_3d_view(w2_data, w2.shape[0], w2.shape[1], w2.shape[2])

    residual_norm = LinearInterpMultiMode3ChannelsResidualNorm(Calpha0, Calpha1,
                                                                  Cw0, Cw1, Cw2)

    if Calpha0 != NULL:
        complex_array_3d_view_free(Calpha0)
    if Calpha1 != NULL:
        complex_array_3d_view_free(Calpha1)
    if Cw0 != NULL:
        complex_array_3d_view_free(Cw0)
    if Cw1 != NULL:
        complex_array_3d_view_free(Cw1)
    if Cw2 != NULL:
        complex_array_3d_view_free(Cw2)

    return residual_norm



#
# Copyright (C) 2019  Sylvain Marsat, John G. Baker
#
#

try:
    # This will fail when lisabeta is imported during the build process,
    # before version.py has been generated.
    from .version import git_hash
    from .version import version as lisabeta_version
except:
    git_hash = 'none'
    lisabeta_version = 'none'

__version__ = lisabeta_version



#
# Copyright (C) 2020 Sylvain Marsat.
#
#


"""
    Python functions for inference with emcee (ensemble sampler, stretch-move)
    with the addition of multimodal jumps in extrinsic parameters.
"""


import time
import numpy as np
import copy
import h5py

import lisabeta.inference.inference as inference

################################################################################
# emcee-like sampler, allowing for a proposal in extra dimensions (mode hopping)
# TODO: include wrapping of phase parameters
# TODO: print info in a user-friendly way
# TODO: allow for checks like m1>m2 in prior_check
# TODO: clean up notation between x_physical and x_extended
# TODO: jump proposal is wasted sometimes by staying in place (50% with 2 modes)
# TODO: have some protection against very wrong prior bounds
# TODO: prior bounds is independent of folding
# TODO: have seen errors with acor of the type:
    # File "/Users/marsat/miniconda3/envs/py36/lib/python3.6/site-packages/acor-1.1.1-py3.6-macosx-10.9-x86_64.egg/acor/acor.py", line 33, in acor
    #     return _acor.acor(np.array(data), maxlag)
    # RuntimeError: D was negative in acor. Can't calculate sigma.

# TODO: make a freely specifiable parameter
# Hardcoded a=2, law 1/sqrt(z) on [1/a, a] = [1/2, 2]
def draw_z():
    u = np.random.uniform()
    return 1./2 * (1.+u)**2

def stretch_move(x, x_ensemble):
    index = np.random.randint(len(x_ensemble))
    x0 = x_ensemble[index]
    z = draw_z()
    y = x0 + z * (x - x0)
    return y, z

def stretch_update(x, lnlike_val, lnpost_val, x_ensemble, lnlikelihood, prior, prior_bounds=None, dim_extra=0, p_extra=0, map_extra=None, proposal_extra=None):

    x_new, lnlike_new, lnpost_new, accept = x, lnlike_val, lnpost_val, False

    dim = len(x) - dim_extra

    x_prop = x.copy()
    # Update in the extra dimensions
    if dim_extra>0 and np.random.binomial(1, p_extra):
        x_prop[dim:] = proposal_extra(x[dim:])
        z = 1.
    # Normal stretch-move update for ordinary dimensions
    else:
        x_prop[:dim], z = stretch_move(x[:dim], x_ensemble[:,:dim])

    # If using extra dimensions, map back to the true point by unfolding the extra dimensions
    if dim_extra>0:
        x_prop_true = map_extra(x_prop)
    else:
        x_prop_true = x_prop

    if inference.prior_check(x_prop[:dim], prior_bounds):
        lnlike_prop = lnlikelihood(x_prop_true)
        prior_prop = prior(x_prop_true)
        lnpost_prop = lnlike_prop + np.log(prior_prop)
        # We assume we always have a sensible lnlike, lnpost
        # if post_prop==0. and post_val==0.:
        #     p = np.fmin(1., z**(dim-1))
        # elif post_val==0.:
        #     p = 1.
        # else:
        p = np.fmin(1., z**(dim-1) * np.exp(lnpost_prop - lnpost_val))
        if np.random.binomial(1, p):
            x_new = x_prop
            lnlike_new = lnlike_prop
            lnpost_new = lnpost_prop
            accept = True

    return x_new, lnlike_new, lnpost_new, accept


# Main multiemcee function
# TODO: print_info, n_iter_info very rudimentary at the moment
def multiemcee_run(lnlikelihood, prior, nwalkers, niter, xini, prior_bounds=None, dim_extra=0, p_extra=0, map_extra=None, proposal_extra=None, seed=None, print_info=False, n_iter_info=10):
    # Random seed
    np.random.seed(seed=seed)
    # Check inputs
    if not nwalkers%2==0:
        raise ValueError('Provide an even number of walkers (split in 2 groups).')
    if not nwalkers==len(xini):
        raise ValueError('nwalkers and xini differ in length.')
    # nw is the number of walkers in each group
    nw = nwalkers//2
    # Initialize walkers with the input xini
    mask = np.full(nwalkers, False, dtype=bool)
    mask[:nw] = True
    np.random.shuffle(mask)
    x_1 = xini[mask]
    x_2 = xini[np.logical_not(mask)]

    # Output arrays for the whole chain, likelihood/posterior values, acceptance
    dim_chain = xini.shape[1]
    chain_shape = (nwalkers, max(1, niter), dim_chain) # niter=0 means we do not iterate
    chain = np.zeros(chain_shape, dtype=float)
    lnlikevals = np.full((nwalkers, niter), 0., dtype=float)
    lnpostvals = np.full((nwalkers, niter), 0., dtype=float)
    acceptvals = np.full((nwalkers, niter), False, dtype=bool)

    # If n_iter=0, we don't even compute likelihoods/priors
    # Simply return the inputs as they are then, with 0s for like, prior
    if niter==0:
        for k in range(nw):
            chain[k,0] = x_1[k]
            chain[nw+k,0] = x_2[k]
        return chain, lnlikevals, lnpostvals, acceptvals

    # Structure to hold the current likelihood, posterior, acceptance
    lnlike_1 = np.zeros(nw, dtype=float)
    lnlike_2 = np.zeros(nw, dtype=float)
    lnpost_1 = np.zeros(nw, dtype=float)
    lnpost_2 = np.zeros(nw, dtype=float)
    accept_1 = np.zeros(nw, dtype=bool)
    accept_2 = np.zeros(nw, dtype=bool)
    dim = dim_chain - dim_extra
    for k in range(nw):
        lnlike_1[k] = lnlikelihood(x_1[k][:dim])
        lnlike_2[k] = lnlikelihood(x_2[k][:dim])
        lnpost_1[k] = np.log(prior(x_1[k][:dim])) + lnlike_1[k]
        lnpost_2[k] = np.log(prior(x_2[k][:dim])) + lnlike_2[k]
        # By convention, set the first acceptance to true
        accept_1[k] = True
        accept_2[k] = True

    # Main loop over iterations
    for i in range(niter):
        if print_info:
            if i%n_iter_info == 0:
                print("n_iter: %d/%d" % (i,niter))
        # Update the 1st group, given the 2nd group
        for k in range(nw):
            x_1[k], lnlike_1[k], lnpost_1[k], accept_1[k] = stretch_update(x_1[k], lnlike_1[k], lnpost_1[k], x_2, lnlikelihood, prior, prior_bounds=prior_bounds, dim_extra=dim_extra, p_extra=p_extra, map_extra=map_extra, proposal_extra=proposal_extra)
            chain[k,i] = x_1[k]
            lnlikevals[k,i] = lnlike_1[k]
            lnpostvals[k,i] = lnpost_1[k]
            acceptvals[k,i] = accept_1[k]
        # Update the 2nd group, given the 1st group
        for k in range(nw):
            x_2[k], lnlike_2[k], lnpost_2[k], accept_2[k] = stretch_update(x_2[k], lnlike_2[k], lnpost_2[k], x_1, lnlikelihood, prior, prior_bounds=prior_bounds, dim_extra=dim_extra, p_extra=p_extra, map_extra=map_extra, proposal_extra=proposal_extra)
            chain[nw+k,i] = x_2[k]
            lnlikevals[nw+k,i] = lnlike_2[k]
            lnpostvals[nw+k,i] = lnpost_2[k]
            acceptvals[nw+k,i] = accept_2[k]

    return chain, lnlikevals, lnpostvals, acceptvals

def unfold_chain(chain, dim_extra, map_extra):
    n = len(chain)
    dim = chain.shape[1] - dim_extra
    chain_unfold = np.zeros((n, dim), dtype=float)
    for i in range(n):
        chain_unfold[i] = map_extra(chain[i])
    return chain_unfold



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Standalone IMRPhenomHM inspiral-merger-ringdown GW waveform model
    for binary black hole coalescences.
"""

import numpy as np
cimport numpy as np
from math import cos

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *

# cdef extern from "IMRPhenomD_internals.h":
#
#     ctypedef struct AmpPhaseFDWaveform:
#         double* freq;
#         double* amp;
#         double* phase;
#         size_t length;
#
#     AmpPhaseFDWaveform* CreateAmpPhaseFDWaveform(
#         size_t length
#     );
#
#     void DestroyAmpPhaseFDWaveform(AmpPhaseFDWaveform* wf);
#
#     ctypedef struct RealVector:
#         double* data;
#         size_t length;
#
#     RealVector* CreateRealVector(
#         size_t length
#     );
#
#     void DestroyRealVector(RealVector* v);

cdef extern from "struct.h":

    ctypedef struct ListAmpPhaseFDMode:
        AmpPhaseFDMode*              hlm;
        int                          l;
        int                          m;
        ListAmpPhaseFDMode*          next;

    ListAmpPhaseFDMode* ListAmpPhaseFDMode_GetMode(
    	   ListAmpPhaseFDMode* list,  # List structure to get this mode from
    	   int l,                     # Mode number l
    	   int m);                    # Mode number m

    ListAmpPhaseFDMode* ListAmpPhaseFDMode_Destroy(
    	   ListAmpPhaseFDMode* list); # List structure to destroy

cdef extern from "IMRPhenomD_internals.h":
    # cdef double _MSUN_SI "MSUN_SI"
    # cdef double _MTSUN_SI "MTSUN_SI"
    # cdef double _MRSUN_SI "MRSUN_SI"
    # cdef double _PC_SI "PC_SI"

    ctypedef struct COMPLEX16FrequencySeries:
        double complex *data;
        char *name;
        long epoch;
        double f0;
        double deltaF;
        size_t length;

    COMPLEX16FrequencySeries *CreateCOMPLEX16FrequencySeries(
        const char *name,
        long epoch,
        double f0,
        double deltaF,
        size_t length
    );

    COMPLEX16FrequencySeries *ResizeCOMPLEX16FrequencySeries(COMPLEX16FrequencySeries *fs, size_t length);
    void DestroyCOMPLEX16FrequencySeries(COMPLEX16FrequencySeries *fs);

cdef extern from "IMRPhenomD_internals.h":
    ctypedef struct ModGRParams:
        double alpha;
    ctypedef struct ExtraParams:
        int use_buggy_LAL_tpeak;
        # int scale_mode_freq;

# # Expose some useful constants hidden in C-space
# MSUN_SI = _MSUN_SI
# MTSUN_SI = _MTSUN_SI
# MRSUN_SI = _MRSUN_SI
# PC_SI = _PC_SI

cdef extern from "IMRPhenomHM.h":
    int IMRPhenomHMGethlmModes(
        ListAmpPhaseFDMode **hlms,   # [out] list of modes, FD amp/phase
        double* fpeak,               # [out] Approximate 22 peak frequency (Hz)
        double* tpeak,               # [out] tf 22 at peak frequency (s)
        double* phipeak,             # [out] phase 22 at peak frequency
        double* fstart,              # [out] Starting frequency (Hz)
        double* tstart,              # [out] tf 22 at starting frequency (s)
        double* phistart,            # [out] phase 22 at starting frequency
        real_vector* freq_22,        # frequency vector for lm in Hz
        real_vector* freq_21,        # frequency vector for lm in Hz
        real_vector* freq_33,        # frequency vector for lm in Hz
        real_vector* freq_32,        # frequency vector for lm in Hz
        real_vector* freq_44,        # frequency vector for lm in Hz
        real_vector* freq_43,        # frequency vector for lm in Hz
        double m1_SI,                # primary mass [kg]
        double m2_SI,                # secondary mass [kg]
        double chi1z,                # aligned spin of primary
        double chi2z,                # aligned spin of secondary
        double distance,             # luminosity distance (Mpc)
        const double phiRef,         # orbital phase at f_ref
        double f_ref,                # reference GW frequency
        const double Deltat,         # Time shift (s) applied a posteriori
        const int scale_freq_hm,     # Scale mode freq by m/2 */
        const ExtraParams* extraparams, # Additional parameters
        const ModGRParams* modgrparams  # Modified GR parameters
    );
    int IMRPhenomHMComputeTimeOfFrequencyModeByMode(
        double* tf22,             # [out] value of t_22 (s)
        double* tf21,             # [out] value of t_21 (s)
        double* tf33,             # [out] value of t_33 (s)
        double* tf32,             # [out] value of t_32 (s)
        double* tf44,             # [out] value of t_44 (s)
        double* tf43,             # [out] value of t_43 (s)
        double f22,              # [in] value of f_22 (Hz)
        double f21,              # [in] value of f_21 (Hz)
        double f33,              # [in] value of f_33 (Hz)
        double f32,              # [in] value of f_32 (Hz)
        double f44,              # [in] value of f_44 (Hz)
        double f43,              # [in] value of f_43 (Hz)
        double m1,                   # primary mass [solar masses]
        double m2,                   # secondary mass [solar masses]
        double chi1z,                   # aligned spin of primary
        double chi2z,                   # aligned spin of secondary
        double distance,                # luminosity distance (Mpc)
        const double phiRef,            # orbital phase at f_ref
        const double fRef_in,                   # reference GW frequency
        const double Deltat,             # Time shift (s) applied a posteriori
        const ExtraParams* extraparams,           # Additional parameters
        const ModGRParams* modgrparams            # Modified GR parameters
    );
    int IMRPhenomHMComputeInverseFrequencyOfTimeModeByMode(
        double* f22,              # [out] value of f_22 (Hz)
        double* f21,              # [out] value of f_21 (Hz)
        double* f33,              # [out] value of f_33 (Hz)
        double* f32,              # [out] value of f_32 (Hz)
        double* f44,              # [out] value of f_44 (Hz)
        double* f43,              # [out] value of f_43 (Hz)
        double tf22,             # [in] value of t_22 (s)
        double tf21,             # [in] value of t_21 (s)
        double tf33,             # [in] value of t_33 (s)
        double tf32,             # [in] value of t_32 (s)
        double tf44,             # [in] value of t_44 (s)
        double tf43,             # [in] value of t_43 (s)
        double f22_estimate,     # [in] guess for the value of f22, will be scaled by m/2
        double t_acc,                 # Target accuracy of t(f) where to stop refining f
        double m1,                   # primary mass [solar masses]
        double m2,                   # secondary mass [solar masses]
        double chi1z,                   # aligned spin of primary
        double chi2z,                   # aligned spin of secondary
        double distance,                # luminosity distance (Mpc)
        const double phiRef,            # orbital phase at f_ref
        const double fRef_in,                   # reference GW frequency
        const double Deltat,             # Time shift (s) applied a posteriori
        const int max_iter,                 # Maximal number of iterations in bisection
        const ExtraParams* extraparams,           # Additional parameters
        const ModGRParams* modgrparams            # Modified GR parameters
    );

modes_PhenomHM = [(2,2), (2,1), (3,3), (3,2), (4,4), (4,3)]

cdef class IMRPhenomHMhlmAmpPhase:
    """ Generate IMRPhenomHM inspiral-merger-ringdown frequency-domain waveform, in Amp/Phase form for hlms.
    """

    cdef ListAmpPhaseFDMode* listhlm # pointer to waveform list of modes
    cdef double fpeak            # Approximate peak frequency (Hz)
    cdef double tpeak            # tf at peak frequency (s)
    cdef double phipeak          # phase 22 at peak frequency
    cdef double fstart           # starting frequency (Hz)
    cdef double tstart           # tf at starting frequency (s)
    cdef double phistart         # phase 22 at starting frequency

    cdef real_vector* Cfreq_22   # pointer to input freqs lm in C structure
    cdef real_vector* Cfreq_21   # pointer to input freqs lm in C structure
    cdef real_vector* Cfreq_33   # pointer to input freqs lm in C structure
    cdef real_vector* Cfreq_32   # pointer to input freqs lm in C structure
    cdef real_vector* Cfreq_44   # pointer to input freqs lm in C structure
    cdef real_vector* Cfreq_43   # pointer to input freqs lm in C structure

    cdef m1                   # Mass of companion 1 (solar masses)
    cdef m2                   # Mass of companion 2 (solar masses)
    cdef chi1                 # Aligned-spin parameter of companion 1 in [-1, 1]
    cdef chi2                 # Aligned-spin parameter of companion 2 in [-1, 1]
    cdef dist                 # Distance of source (Mpc)
    # cdef tref                 # Time at fref_for_tref (s)
    cdef phiref               # Orbital phase at fref_for_phiref (rad)
    cdef fref                 # ref. frequency (Hz) for tref
    # cdef fref_for_tref        # ref. frequency (Hz) for tref
    # cdef fref_for_phiref      # ref. frequency (Hz) for phiref
    # cdef force_phiref_fref    # Flag forcing phiref at fref after adjusting tref
    cdef Deltat               # Time shift (s) applied a posteriori
    cdef f_min, f_max
    cdef length
    # cdef freq, amp, phase

    cdef ModGRParams* Cmodgrparams # pointer to struct with mod GR params
    cdef ModGRParams Cmodgrparams_struct

    cdef ExtraParams* Cextraparams # pointer to struct with extra params
    cdef ExtraParams Cextraparams_struct

    cdef public object modes  # List of modes (l,m)
    cdef hlm                  # Dictionary of modes
    cdef scale_freq_hm        # Scale mode freq by m/2

    # # Used for the f(t), t(f) functions
    # cdef double tf22            # Time of frequency for mode lm (s)
    # cdef double tf21            # Time of frequency for mode lm (s)
    # cdef double tf33            # Time of frequency for mode lm (s)
    # cdef double tf32            # Time of frequency for mode lm (s)
    # cdef double tf44            # Time of frequency for mode lm (s)
    # cdef double tf43            # Time of frequency for mode lm (s)
    # cdef double f22             # Frequency of time for mode lm (Hz)
    # cdef double f21             # Frequency of time for mode lm (Hz)
    # cdef double f33             # Frequency of time for mode lm (Hz)
    # cdef double f32             # Frequency of time for mode lm (Hz)
    # cdef double f44             # Frequency of time for mode lm (Hz)
    # cdef double f43             # Frequency of time for mode lm (Hz)

    # modes_PhHM = [(2,2), (2,1), (3,3), (3,2), (4,4), (4,3)]

    def __init__(self,
                 freq,
                 m1, m2, chi1, chi2, dist,
                 phiref=0., fref=0., Deltat=0.,
                 modes=modes_PhenomHM, scale_freq_hm=True,
                 extra_params=None, mod_gr_params=None):
                 # tref=0., fref_for_tref=0., fref_for_phiref=0.,
                 # force_phiref_fref=True, Deltat=None):
        """Constructor
        Args:
          freq              # Frequencies (Hz) on which to evaluate waveform
                              If numpy 1D array, interpreted as f_22
                              higher modes are returned on frequencies m/2*freq
                              if scale_freq_hm is set to True
                              If dict, dict of np array for each f_lm
                              (NOTE: if dict, scale_freq_hm ignored)
          m1                # Mass of companion 1 (solar masses)
          m2                # Mass of companion 2 (solar masses)
          chi1              # Aligned-spin parameter of companion 1 in [-1,1]
          chi2              # Aligned-spin parameter of companion 2 in [-1,1]
          dist              # Luminosity distance of source (Mpc)
        Keyword args:
          fref                # ref. frequency (Hz)
          phiref              # Orbital phase at fref (rad)
          Deltat              # Time shift (s) applied a posteriori
          modes               # List of modes (l,m) to generate
          scale_freq_hm       # Scale mode freq by m/2 (default True)
                                if input freq is dict, ignored
          extra_params        # Dictionary of additional parameters
          |- use_buggy_LAL_tpeak # Reproduce bug in tpeak alignment in LAL
                                   (default False)
          mod_gr_params       # Dictionary of modified GR parameters (NOT ready yet)
        """
          # tref                # Time at fref_for_tref (s)
          # phiref              # Orbital phase at fref_for_phiref (rad)
          # fref_for_tref       # ref. frequency (Hz) for tref
          # fref_for_phiref     # ref. frequency (Hz) for phiref
          # force_phiref_fref   # Flag forcing phiref at fref after adjusting tref
          # Deltat              # Time shift (s) applied a posteriori

        # Input frequencies can be either a 1d np array for f_22
        # or a dict with a 1d np array for each f_lm
        # TODO: very clunky, need to find how to pass list structures to cython
        cdef np.ndarray[np.float_t, ndim=1] freq_22 = None # Frequencies for the mode lm
        cdef np.ndarray[np.float_t, ndim=1] freq_21 = None # Frequencies for the mode lm
        cdef np.ndarray[np.float_t, ndim=1] freq_33 = None # Frequencies for the mode lm
        cdef np.ndarray[np.float_t, ndim=1] freq_32 = None # Frequencies for the mode lm
        cdef np.ndarray[np.float_t, ndim=1] freq_44 = None # Frequencies for the mode lm
        cdef np.ndarray[np.float_t, ndim=1] freq_43 = None # Frequencies for the mode lm
        if isinstance(freq, np.ndarray):
            freq_22 = freq
        elif isinstance(freq, dict):
            scale_freq_hm = False # Giving explicit mode-by-mode freq supersedes this option
            freq_22 = freq.get((2,2), None)
            freq_21 = freq.get((2,1), None)
            freq_33 = freq.get((3,3), None)
            freq_32 = freq.get((3,2), None)
            freq_44 = freq.get((4,4), None)
            freq_43 = freq.get((4,3), None)
        else:
            raise ValueError('freq must be either a numpy array (for f_22)'\
                            +' or a dict of arrays for the f_lm.')

        # arguments are checked in the C waveform generator
        self.fpeak = 0.
        self.tpeak = 0.
        self.phipeak = 0.
        self.fstart = 0.
        self.tstart = 0.
        self.phistart = 0.
        self.m1 = m1
        self.m2 = m2
        self.chi1 = chi1
        self.chi2 = chi2
        self.dist = dist
        #self.tref = tref
        self.phiref = phiref
        self.fref = fref
        # self.fref_for_tref = fref_for_tref
        # self.fref_for_phiref = fref_for_phiref
        # self.force_phiref_fref = force_phiref_fref
        if Deltat is None:
            Deltat = 0.
        self.Deltat = Deltat
        self.f_min = freq_22[0]
        self.f_max = freq_22[-1]
        self.length = len(freq_22)
        self.modes = modes
        self.scale_freq_hm = scale_freq_hm

        self.listhlm = NULL

        # Build a real_vector representation of the input numpy array data
        # TODO: very clunky, need to find how to pass list structures to cython
        self.Cfreq_22 = NULL
        cdef double* freq_22_data = NULL
        if freq_22 is not None:
            if not freq_22.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_22_data = <double *> &freq_22[0]
            self.Cfreq_22 = real_vector_view(freq_22_data, freq_22.shape[0])
        self.Cfreq_21 = NULL
        cdef double* freq_21_data = NULL
        if freq_21 is not None:
            if not freq_21.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_21_data = <double *> &freq_21[0]
            self.Cfreq_21 = real_vector_view(freq_21_data, freq_21.shape[0])
        self.Cfreq_33 = NULL
        cdef double* freq_33_data = NULL
        if freq_33 is not None:
            if not freq_33.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_33_data = <double *> &freq_33[0]
            self.Cfreq_33 = real_vector_view(freq_33_data, freq_33.shape[0])
        self.Cfreq_32 = NULL
        cdef double* freq_32_data = NULL
        if freq_32 is not None:
            if not freq_32.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_32_data = <double *> &freq_32[0]
            self.Cfreq_32 = real_vector_view(freq_32_data, freq_32.shape[0])
        self.Cfreq_44 = NULL
        cdef double* freq_44_data = NULL
        if freq_44 is not None:
            if not freq_44.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_44_data = <double *> &freq_44[0]
            self.Cfreq_44 = real_vector_view(freq_44_data, freq_44.shape[0])
        self.Cfreq_43 = NULL
        cdef double* freq_43_data = NULL
        if freq_43 is not None:
            if not freq_43.flags['C_CONTIGUOUS']:
                raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
            freq_43_data = <double *> &freq_43[0]
            self.Cfreq_43 = real_vector_view(freq_43_data, freq_43.shape[0])

        # cdef int Cforce_phiref_fref = <int> self.force_phiref_fref
        cdef int Cscale_freq_hm = <int> self.scale_freq_hm

        # Build C structure holding additional params from input python dict
        # if input is None, C struct will stay NULL and code will use default
        self.Cextraparams = NULL
        if extra_params is not None:
            self.Cextraparams_struct.use_buggy_LAL_tpeak = \
                                  extra_params.get('use_buggy_LAL_tpeak', False)
            self.Cextraparams = &self.Cextraparams_struct

        # Build C structure holding modified GR params from input python dict
        # if input is None, C struct will stay NULL and be ignored in C code
        self.Cmodgrparams = NULL
        if mod_gr_params is not None:
            self.Cmodgrparams_struct.alpha = mod_gr_params['alpha'] #PLACEHOLDER
            self.Cmodgrparams = &self.Cmodgrparams_struct

        # ModGRParams not fully implemented yet, notably for setup stage
        if not (mod_gr_params is None):
            raise ValueError('ModGRParams for PhenomHM not fully implemented yet.')

        ret = IMRPhenomHMGethlmModes(
            &self.listhlm,
            &self.fpeak,
            &self.tpeak,
            &self.phipeak,
            &self.fstart,
            &self.tstart,
            &self.phistart,
            self.Cfreq_22,
            self.Cfreq_21,
            self.Cfreq_33,
            self.Cfreq_32,
            self.Cfreq_44,
            self.Cfreq_43,
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.phiref,
            self.fref,
            self.Deltat,
            Cscale_freq_hm,
            self.Cextraparams,
            self.Cmodgrparams
        );

        if ret != 0:
            raise ValueError("Call to IMRPhenomHMGethlmModes() failed.")

        # Read the modes in a dictionary
        # Direct copy of C double array to numpy via a MemoryView
        # TODO: the AmpPhaseFDMode structures allow for a complex amplitude
        # and for different frequency vectors for the amplitude and phase
        # for now, we simply keep a simple freq, amp, phase structure
        cdef AmpPhaseFDMode* Chlm = NULL
        self.hlm = {}
        cdef double[::1] view_freq_amp
        cdef double[::1] view_amp_real
        # cdef double[::1] view_amp_imag
        # cdef double[::1] view_freq_phase
        cdef double[::1] view_phase
        cdef double[::1] view_tf
        for lm in self.modes:
            (l,m) = (lm[0], lm[1])
            self.hlm[lm] = {}
            if not lm in self.modes:
                raise ValueError('Mode not allowed: (%d, %d)' % (l, m))
            Chlm = ListAmpPhaseFDMode_GetMode(self.listhlm, l, m).hlm
            view_freq_amp = \
                        <(double)[:Chlm.freq_amp.size]> Chlm.freq_amp.data
            view_amp_real = \
                        <(double)[:Chlm.amp_real.size]> Chlm.amp_real.data
            # view_amp_imag = \
            #             <(double)[:Chlm.amp_imag.size]> Chlm.amp_imag.data
            # view_freq_phase = \
            #             <(double)[:Chlm.freq_phase.size]> Chlm.freq_phase.data
            view_phase = \
                        <(double)[:Chlm.phase.size]> Chlm.phase.data
            view_tf = \
                        <(double)[:Chlm.tf.size]> Chlm.tf.data
            # self.hlm[lm]['freq_amp'] = np.asarray(view_freq_amp)
            # self.hlm[lm]['amp_real'] = np.asarray(view_amp_real)
            # self.hlm[lm]['amp_imag'] = np.asarray(view_amp_imag)
            # self.hlm[lm]['freq_phase'] = np.asarray(view_freq_phase)
            # self.hlm[lm]['phase'] = np.asarray(view_phase)
            self.hlm[lm]['freq'] = np.asarray(view_freq_amp)
            self.hlm[lm]['amp'] = np.asarray(view_amp_real)
            self.hlm[lm]['phase'] = np.asarray(view_phase)
            self.hlm[lm]['tf'] = np.asarray(view_tf)

    def __dealloc__(self):
        """Destructor
        """
        if self.listhlm != NULL:
            ListAmpPhaseFDMode_Destroy(self.listhlm)

    # NOTE: to return a copy of the modes, copy.deepcopy is noticeably slower
    # TODO: the AmpPhaseFDMode structures allow for a complex amplitude
    # and for different frequency vectors for the amplitude and phase
    # for now, we simply keep a simple freq, amp, phase structure
    def get_waveform(self):
        hlm = {}
        for lm in self.modes:
            (l,m) = (lm[0], lm[1])
            hlm[lm] = {}
            # hlm[lm]['freq_amp'] = np.copy(self.hlm[lm]['freq_amp'])
            # hlm[lm]['amp_real'] = np.copy(self.hlm[lm]['amp_real'])
            # hlm[lm]['amp_imag'] = np.copy(self.hlm[lm]['amp_imag'])
            # hlm[lm]['freq_phase'] = np.copy(self.hlm[lm]['freq_phase'])
            # hlm[lm]['phase'] = np.copy(self.hlm[lm]['phase'])
            hlm[lm]['freq'] = np.copy(self.hlm[lm]['freq'])
            hlm[lm]['amp'] = np.copy(self.hlm[lm]['amp'])
            hlm[lm]['phase'] = np.copy(self.hlm[lm]['phase'])
            hlm[lm]['tf'] = np.copy(self.hlm[lm]['tf'])
        return hlm

    def get_fpeak(self):
        return self.fpeak

    def get_tpeak(self):
        return self.tpeak

    def get_phipeak(self):
        return self.phipeak

    def get_fstart(self):
        return self.fstart

    def get_tstart(self):
        return self.tstart

    def get_phistart(self):
        return self.phistart

    def compute_toff_lm(self, f_lm):
        cdef double tf22 = 0.
        cdef double tf21 = 0.
        cdef double tf33 = 0.
        cdef double tf32 = 0.
        cdef double tf44 = 0.
        cdef double tf43 = 0.
        t = IMRPhenomHMComputeTimeOfFrequencyModeByMode(
            &tf22,
            &tf21,
            &tf33,
            &tf32,
            &tf44,
            &tf43,
            f_lm[(2,2)],
            f_lm[(2,1)],
            f_lm[(3,3)],
            f_lm[(3,2)],
            f_lm[(4,4)],
            f_lm[(4,3)],
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.phiref,
            self.fref,
            self.Deltat,
            self.Cextraparams,
            self.Cmodgrparams
        );
        tf_lm = {}
        tf_lm[(2,2)] = tf22
        tf_lm[(2,1)] = tf21
        tf_lm[(3,3)] = tf33
        tf_lm[(3,2)] = tf32
        tf_lm[(4,4)] = tf44
        tf_lm[(4,3)] = tf43
        return tf_lm

    def compute_foft_lm(self, tf_lm, f22_estimate, t_acc, max_iter=100):
        cdef double f22 = 0.
        cdef double f21 = 0.
        cdef double f33 = 0.
        cdef double f32 = 0.
        cdef double f44 = 0.
        cdef double f43 = 0.
        cdef int Cmax_iter = <int> max_iter
        f = IMRPhenomHMComputeInverseFrequencyOfTimeModeByMode(
            &f22,
            &f21,
            &f33,
            &f32,
            &f44,
            &f43,
            tf_lm[(2,2)],
            tf_lm[(2,1)],
            tf_lm[(3,3)],
            tf_lm[(3,2)],
            tf_lm[(4,4)],
            tf_lm[(4,3)],
            f22_estimate,
            t_acc,
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.phiref,
            self.fref,
            self.Deltat,
            Cmax_iter,
            self.Cextraparams,
            self.Cmodgrparams
        );
        f_lm = {}
        f_lm[(2,2)] = f22
        f_lm[(2,1)] = f21
        f_lm[(3,3)] = f33
        f_lm[(3,2)] = f32
        f_lm[(4,4)] = f44
        f_lm[(4,3)] = f43

        return f_lm



from distutils.core import setup, Extension
from Cython.Build import cythonize
import numpy
from distutils.sysconfig import get_python_lib
import sys


gsl_prefix="/usr"
lisabeta_include="../.."

argv_replace = []
for arg in sys.argv:
    if arg.startswith('--gsl-prefix='):
        gsl_prefix = arg.split('=', 1)[1]
    else:
        argv_replace.append(arg)
sys.argv = argv_replace

gsl_lib = gsl_prefix+"/lib"
gsl_include = gsl_prefix+"/include"

extensions=[
    Extension("pyIMRPhenomD",
              sources=["pyIMRPhenomD.c", "IMRPhenomD.c", "IMRPhenomD_internals.c"],
              include_dirs = [numpy.get_include(),lisabeta_include,gsl_include],
              language="c",
              extra_compile_args = ["-std=c99", "-O3"],
              libraries=["gsl", "gslcblas"],
              library_dirs=[gsl_lib],
    )
]
setup(
    ext_modules = extensions
)

setup(
    ext_modules = cythonize("pyIMRPhenomD.pyx")
)




#!/usr/bin/env python
# Copyright (C) 2019 Sylvain Marsat, John G. Baker
#

"""
setup.py file for lisabeta package
"""

from __future__ import print_function

import sys, os, subprocess, shutil

from distutils.errors import DistutilsError
from distutils.command.clean import clean as _clean

from setuptools.command.install import install as _install
from setuptools import Extension, setup, Command
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools import find_packages

from Cython.Build import cythonize

import numpy

# Classifiers for distribution of package
classifiers = [
    'Programming Language :: Python',
    'Programming Language :: Cython',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Topic :: Scientific/Engineering :: Physics',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
]

requires = []
setup_requires = ['numpy',]
install_requires = setup_requires + [
                      'cython',
                      'scipy',
                      'matplotlib',
                      'h5py',
                      'astropy'
                      ]

# def find_files(dirname, relpath=None):
#     def find_paths(dirname):
#         items = []
#         for fname in os.listdir(dirname):
#             path = os.path.join(dirname, fname)
#             if os.path.isdir(path):
#                 items += find_paths(path)
#             elif not path.endswith(".py") and not path.endswith(".pyc"):
#                 items.append(path)
#         return items
#     items = find_paths(dirname)
#     if relpath is None:
#         relpath = dirname
#     return [os.path.relpath(path, relpath) for path in items]

# class cbuild_ext(_build_ext):
#     def run(self):
#         import pkg_resources
#
#         # At this point we can be sure pip has already installed numpy
#         numpy_incl = pkg_resources.resource_filename('numpy', 'core/include')
#
#         for ext in self.extensions:
#             if (hasattr(ext, 'include_dirs') and
#                     numpy_incl not in ext.include_dirs):
#                 ext.include_dirs.append(numpy_incl)
#
#         _build_ext.run(self)

# # write versioning info
# def get_version_info():
#     """Get VCS info and write version info to version.py
#     """
#     from pycbc import _version_helper
#
#     class vdummy(object):
#         def __getattr__(self, attr):
#             return ''
#
#     # If this is a pycbc git repo always populate version information using GIT
#     try:
#         vinfo = _version_helper.generate_git_version_info()
#     except:
#         vinfo = vdummy()
#         vinfo.version = '1.13.dev7'
#         vinfo.release = 'False'
#
#     with open('pycbc/version.py', 'w') as f:
#         f.write("# coding: utf-8\n")
#         f.write("# Generated by setup.py for PyCBC on %s.\n\n"
#                 % vinfo.build_date)
#
#         # print general info
#         f.write('version = \'%s\'\n' % vinfo.version)
#         f.write('date = \'%s\'\n' % vinfo.date)
#         f.write('release = %s\n' % vinfo.release)
#         f.write('last_release = \'%s\'\n' % vinfo.last_release)
#
#         # print git info
#         f.write('\ngit_hash = \'%s\'\n' % vinfo.hash)
#         f.write('git_branch = \'%s\'\n' % vinfo.branch)
#         f.write('git_tag = \'%s\'\n' % vinfo.tag)
#         f.write('git_author = \'%s\'\n' % vinfo.author)
#         f.write('git_committer = \'%s\'\n' % vinfo.committer)
#         f.write('git_status = \'%s\'\n' % vinfo.status)
#         f.write('git_builder = \'%s\'\n' % vinfo.builder)
#         f.write('git_build_date = \'%s\'\n' % vinfo.build_date)
#         f.write('git_verbose_msg = """Version: %s\n'
#                 'Branch: %s\n'
#                 'Tag: %s\n'
#                 'Id: %s\n'
#                 'Builder: %s\n'
#                 'Build date: %s\n'
#                 'Repository status is %s"""\n' %(
#                                                vinfo.version,
#                                                vinfo.branch,
#                                                vinfo.tag,
#                                                vinfo.hash,
#                                                vinfo.builder,
#                                                vinfo.build_date,
#                                                vinfo.status))
#         f.write('from pycbc._version import *\n')
#         version = vinfo.version
#
#     from pycbc import version
#     version = version.version
#     return version

# class build_docs(Command):
#     user_options = []
#     description = "Build the documentation pages"
#     def initialize_options(self):
#         pass
#     def finalize_options(self):
#         pass
#     def run(self):
#         subprocess.check_call("cd docs; cp Makefile.std Makefile; cp conf_std.py conf.py; sphinx-apidoc "
#                               " -o ./ -f -A 'PyCBC dev team' -V '0.1' ../pycbc && make html",
#                             stderr=subprocess.STDOUT, shell=True)

# class build_gh_pages(Command):
#     user_options = []
#     description = "Build the documentation pages for GitHub"
#     def initialize_options(self):
#         pass
#     def finalize_options(self):
#         pass
#     def run(self):
#         subprocess.check_call("mkdir -p _gh-pages/latest && touch _gh-pages/.nojekyll && "
#                               "cd docs; cp Makefile.gh_pages Makefile; cp conf_std.py conf.py; sphinx-apidoc "
#                               " -o ./ -f -A 'PyCBC dev team' -V '0.1' ../pycbc && make html",
#                             stderr=subprocess.STDOUT, shell=True)

# cmdclass = { 'build_docs' : build_docs,
#              'build_gh_pages' : build_gh_pages,
#              'clean' : clean,
#              'build_ext':cbuild_ext
#             }

#extras_require = {'cuda': ['pycuda>=2015.1', 'scikit-cuda']}

# do the actual work of building the package
#VERSION = get_version_info()

argv_replace = []
constants = ''
gsl_prefix = '/usr'
force_recompile=False
for arg in sys.argv:
    if arg.startswith('--with-gsl='):
        gsl_prefix = arg.split('=', 1)[1]
    elif arg.startswith('--with-constants='):
        constants = arg.split('=', 1)[1]
    elif arg=='--force':
        force_recompile = True
    else:
        argv_replace.append(arg)
sys.argv = argv_replace

# If using LDC constants, define a macro to pass to the compiler
if constants=='' or constants=='LAL':
    Dconstants = '-DNOP'
elif constants=='LDC':
    Dconstants = '-DLDCCONSTANTS'
elif constants=='LISA':
    Dconstants = '-DLISACONSTANTS'

#Compiler
#os.environ["CC"] = "/opt/local/bin/gcc-mp-8"
#os.environ["CXX"] = "/opt/local/bin/g++-mp-8"

# Extensions
lib_gsl_dir = gsl_prefix+"/lib"
include_gsl_dir = gsl_prefix+"/include"
# Note: we use only one list of include dirs with everything needed in it
#extra_compile_args = [Dconstants]
extra_link_args= []
#if(0):
    #compiler for memory debugging:
    #Note: to get this to run I need to execute (for macports python 3.7):
    # DYLD_INSERT_LIBRARIES=/Applications/Xcode10.1.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/lib/clang/10.0.0/lib/darwin/libclang_rt.asan_osx_dynamic.dylib /opt/local/Library/Frameworks/Python.framework/Versions/3.7/Resources/Python.app/Contents/MacOS/Python
    # extra_compile_args += [ "-fsanitize=address","-fno-omit-frame-pointer", "-Wno-format-security"]
    # extra_link_args += ["-fsanitize=address"]
extra_compile_args = ['-std=c99']
all_include_dirs = [numpy.get_include(), include_gsl_dir, "lisabeta/",
                    "lisabeta/struct", "lisabeta/tools", "lisabeta/lisa"]
all_libraries = ["gsl", "gslcblas"]
all_library_dirs = [lib_gsl_dir]
def func_extension(ext, sources=[]):
    return Extension(ext,
              sources=sources,
              include_dirs=all_include_dirs,
              language="c",
              extra_compile_args=extra_compile_args,
              #extra_link_args=extra_link_args,
              libraries=all_libraries,
              library_dirs=all_library_dirs)
extensions = [
    func_extension("lisabeta.pyconstants",
      sources=["lisabeta/pyconstants.pyx"]),
    func_extension("lisabeta.struct.pystruct",
      sources=["lisabeta/struct/pystruct.pyx",
      "lisabeta/struct/struct.c"]),
    func_extension("lisabeta.tools.pytools",
      sources=["lisabeta/tools/pytools.pyx",
      "lisabeta/tools/tools.c",
      "lisabeta/struct/struct.c",
      "lisabeta/tools/spline.c"]),
    func_extension("lisabeta.tools.pyspline",
      sources=["lisabeta/tools/pyspline.pyx",
      "lisabeta/tools/spline.c",
      "lisabeta/struct/struct.c",
      "lisabeta/tools/tools.c"]),
    func_extension("lisabeta.tools.pyoverlap",
      sources=["lisabeta/tools/pyoverlap.pyx",
      "lisabeta/tools/spline.c",
      "lisabeta/struct/struct.c",
      "lisabeta/tools/tools.c",
      "lisabeta/tools/fresnel.c",
      "lisabeta/tools/overlap.c"]),
    func_extension("lisabeta.lisa.pyresponse",
      sources=["lisabeta/lisa/pyresponse.pyx",
      "lisabeta/lisa/LISAresponse.c",
      "lisabeta/lisa/LISAgeometry.c",
      "lisabeta/tools/spline.c",
      "lisabeta/struct/struct.c",
      "lisabeta/tools/tools.c"]),
    func_extension("lisabeta.lisa.pyLISAnoise",
      sources=["lisabeta/lisa/pyLISAnoise.pyx",
      "lisabeta/lisa/LISAnoise.c",
      "lisabeta/lisa/LISAgeometry.c",
      "lisabeta/struct/struct.c"]),
    func_extension("lisabeta.waveforms.bbh.pyIMRPhenomD",
      sources=["lisabeta/waveforms/bbh/pyIMRPhenomD.pyx",
      "lisabeta/waveforms/bbh/IMRPhenomD.c",
      "lisabeta/waveforms/bbh/IMRPhenomD_internals.c",
      "lisabeta/waveforms/bbh/IMRPhenomUtils.c",
      "lisabeta/waveforms/bbh/IMRPhenomInternalUtils.c",
      "lisabeta/struct/struct.c"]),
    func_extension("lisabeta.waveforms.bbh.pyIMRPhenomHM",
      sources=["lisabeta/waveforms/bbh/pyIMRPhenomHM.pyx",
      "lisabeta/waveforms/bbh/IMRPhenomHM.c",
      "lisabeta/waveforms/bbh/IMRPhenomD_internals.c",
      "lisabeta/waveforms/bbh/IMRPhenomUtils.c",
      "lisabeta/waveforms/bbh/IMRPhenomInternalUtils.c",
      "lisabeta/waveforms/bbh/RingdownCW.c",
      "lisabeta/struct/struct.c"]),
    func_extension("lisabeta.waveforms.bbh.pyEOBNRv2HMROM",
      sources=["lisabeta/waveforms/bbh/pyEOBNRv2HMROM.pyx",
      "lisabeta/waveforms/bbh/EOBNRv2HMROM.c",
      "lisabeta/waveforms/bbh/EOBNRv2HMROMstruct.c",
      "lisabeta/struct/struct.c"])
]
ext_modules = cythonize(extensions,force=force_recompile)


setup(
    name = 'lisabeta',
    version = '0.1.0',
    description = 'Library to analyze gravitational-wave signals in LISA and LIGO-type detectors.',
    #long_description = open('descr.rst').read(),
    author = 'Sylvain Marsat and John Baker and ...',
    author_email = 'marsat@apc.in2p3.fr',
    #url = 'http://www.pycbc.org/',
    #download_url = 'https://github.com/gwastro/pycbc/tarball/v%s' % VERSION,
    keywords = ['lisa', 'ligo', 'physics', 'gravity', 'signal processing', 'gravitational waves'],
    #cmdclass = cmdclass,
    setup_requires = setup_requires,
    #extras_require = extras_require,
    install_requires = install_requires,
    #scripts  = find_files('bin', relpath='./') + ['tools/einsteinathome/pycbc_build_eah.sh'],
    packages = find_packages(),
    #package_data = {'pycbc.workflow': find_files('pycbc/workflow'),
    #                'pycbc.results': find_files('pycbc/results'),
    #                'pycbc.tmpltbank': find_files('pycbc/tmpltbank')},
    ext_modules = ext_modules,
    classifiers = classifiers
)
